# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


import os.path
import logging
from typing import Any

from xpra.util.env import envbool

SILENCE_COMTYPES = envbool("XPRA_SILENCE_COMTYPES", True)
if SILENCE_COMTYPES:
    logging.getLogger("comtypes").setLevel(logging.INFO)
COMTYPES_NOGENDIR = envbool("XPRA_COMTYPES_NOGENDIR", False)

COMTYPES_ENABLED = envbool("XPRA_COMTYPES", True)


def comtypes_init() -> None:
    if not COMTYPES_ENABLED:
        raise RuntimeError("comtypes is disabled")
    # pylint: disable=import-outside-toplevel
    from comtypes import client
    if COMTYPES_NOGENDIR:
        client.gen_dir = None
    from comtypes import CoInitialize
    CoInitialize()


def find_tlb_file(filename: str = "DirectShow.tlb") -> str:
    # try to load tlb files from various directories,
    # depending on how xpra was packaged, installed locally,
    # or even run from the source directory:
    from xpra.platform.paths import get_app_dir  # pylint: disable=import-outside-toplevel
    app_dir = get_app_dir()
    dirs = [
        app_dir,
        os.path.join(app_dir, "lib", "tlb"),
        os.path.join(app_dir, "win32"),
        os.path.join(app_dir, "share", "xpra"),
        os.path.join(os.environ.get("MINGW_PREFIX", ""), "share", "xpra"),
    ]
    # ie: "DirectShow.tlb" -> "XPRA_DIRECTSHOW_TLB"
    env_name = "XPRA_" + filename.replace(".", "_").upper()
    filenames = [os.environ.get(env_name)] + [os.path.join(d, filename) for d in dirs]
    for f in filenames:
        if f and os.path.exists(f):
            return f
    return ""


class QuietenLogging:

    def __init__(self, *_args):
        self.loggers = [logging.getLogger(x) for x in ("comtypes.client._code_cache", "comtypes.client._generate")]
        self.saved_levels = [x.getEffectiveLevel() for x in self.loggers]
        self._generate: Any = None

    def __enter__(self):
        if not SILENCE_COMTYPES:
            return
        for logger in self.loggers:
            logger.setLevel(logging.WARNING)
        self.verbose = None
        from comtypes import client  # pylint: disable=import-outside-toplevel
        gen = getattr(client, "_generate", None)
        self._generate = gen
        if gen:
            self.verbose = getattr(gen, "__verbose__", None)
            if self.verbose is not None:
                gen.__verbose__ = False

    def __exit__(self, *_args):
        if not SILENCE_COMTYPES:
            return
        gen = self._generate
        if gen and self.verbose is not None:
            gen.__verbose__ = self.verbose
        for i, logger in enumerate(self.loggers):
            logger.setLevel(self.saved_levels[i])


class CIMV2_Query(QuietenLogging):

    def __init__(self, query):
        self.query = query
        super().__init__()

    def __enter__(self):
        super().__enter__()
        from comtypes.client import CreateObject  # pylint: disable=import-outside-toplevel
        o = CreateObject("WbemScripting.SWbemLocator")
        s = o.ConnectServer(".", "root\\cimv2")
        return s.ExecQuery(self.query)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import hmac
import hashlib
import unittest

class TestHMAC(unittest.TestCase):

    def test_hardcoded(self):
        password = b"71051d81d27745b59c1c56c6e9046c19697e452453e04aa5abbd52c8edc8c232"
        salt = b"99ea464f-7117-4e38-95b3-d3aa80e7b806"
        try:
            hmac_hash = hmac.HMAC(password, salt, digestmod=hashlib.md5)
            hd = hmac_hash.hexdigest()
            ed = "dc26a074c9378b1b5735a27563320a26"
        except ValueError:
            hmac_hash = hmac.HMAC(password, salt, digestmod=hashlib.sha1)
            hd = hmac_hash.hexdigest()
            ed = "5529d27aef1b7420fb9d696f1c04eaad5dcc1515"
        assert hd == ed, "expected digest %s but got %s" % (ed, hd)

def main():
    unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>

import sys
import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, Gdk	#pylint: disable=wrong-import-position @UnresolvedImport


class TestForm(object):

	def	__init__(self):
		self.window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
		self.window.connect("destroy", Gtk.main_quit)
		self.window.set_default_size(320, 200)
		self.window.set_border_width(20)
		self.window.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)

		btn = Gtk.Button(label="hello")
		btn.connect("button_press_event", self.button_clicked)

		self.window.add(btn)
		self.window.show_all()

	def button_clicked(self, widget, event):
		print("button_clicked("+str(widget)+", "+str(event)+")")
		menu = Gtk.Menu()
		menu.append(Gtk.MenuItem(label="Foo"))
		menu.append(Gtk.MenuItem(label="Bar"))
		menu.show_all()
		menu.popup(None, None, None, None, event.button, event.time)


def main():
	TestForm()
	Gtk.main()


if __name__ == "__main__":
	main()
	sys.exit(0)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import unittest

from unit.test_util import silence_error, LoggerSilencer
from xpra.server import background_worker
from xpra.server.background_worker import get_worker, add_work_item, stop_worker, Worker_Thread


class BackgroundWorkerTest(unittest.TestCase):

    def test_run(self):
        assert get_worker(False) is None
        w = get_worker()
        assert repr(w)

        def error_item():
            raise Exception("work item test error")

        with silence_error(background_worker):
            add_work_item(error_item)
            time.sleep(0.1)
        # add the same item twice, with "no-duplicates"
        # (should only get added once)
        ndc = []

        def nodupe():
            ndc.append(True)

        w.add(nodupe, False)
        w.add(nodupe, False)
        time.sleep(1)
        with LoggerSilencer(background_worker):
            # trigger the warning with more than 10 items:
            def slow_item():
                time.sleep(1)

            for _ in range(12):
                w.add(slow_item)
            stop_worker()
            stop_worker(True)
        # no-op:
        stop_worker(True)
        # let the worker print its messages:
        time.sleep(1)
        assert len(ndc) == 1, "nodupe item should have been run once only, got %i" % (len(ndc),)

    def test_normal_stop(self):
        w = Worker_Thread()
        w.start()
        # noinspection PyTypeChecker
        w.add(None)
        time.sleep(1)
        assert w.exit is True


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic

from xpra.gtk.gobject import n_arg_signal, one_arg_signal
from xpra.clipboard.core import ClipboardProxyCore, TEXT_TARGETS, ClipboardCallback
from xpra.clipboard.timeout import ClipboardTimeoutHelper
from xpra.os_util import gi_import
from xpra.util.str_fn import Ellipsizer
from xpra.util.env import envint
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GObject = gi_import("GObject")

log = Logger("clipboard")

BLOCK_DELAY = envint("XPRA_CLIPBOARD_BLOCK_DELAY", 5)


class GTK_Clipboard(ClipboardTimeoutHelper):

    def __repr__(self):
        return "GTK_Clipboard"

    def make_proxy(self, selection):
        proxy = GTKClipboardProxy(selection)
        proxy.set_want_targets(self._want_targets)
        proxy.set_direction(self.can_send, self.can_receive)
        proxy.connect("send-clipboard-token", self._send_clipboard_token_handler)
        proxy.connect("send-clipboard-request", self._send_clipboard_request_handler)
        return proxy


class GTKClipboardProxy(ClipboardProxyCore, GObject.GObject):
    __gsignals__ = {
        "send-clipboard-token": one_arg_signal,
        "send-clipboard-request": n_arg_signal(2),
    }

    def __init__(self, selection="CLIPBOARD"):
        ClipboardProxyCore.__init__(self, selection)
        GObject.GObject.__init__(self)
        self._owner_change_embargo = 0.0
        self._want_targets = False
        display = Gdk.Display.get_default()
        if not display:
            log.warn(f"Warning: no display, cannot access the {selection} clipboard")
            return
        self.clipboard = Gtk.Clipboard.get(Gdk.Atom.intern(selection, False))
        self.clipboard.connect("owner-change", self.owner_change)

    def __repr__(self):
        return "GTKClipboardProxy(%s)" % self._selection

    def got_token(self, targets, target_data=None, claim=True, synchronous_client=False) -> None:
        # the remote end now owns the clipboard
        self.cancel_emit_token()
        if not self._enabled:
            return
        self._got_token_events += 1
        log("got token, selection=%s, targets=%s, target data=%s, claim=%s, synchronous_client=%s, can-receive=%s",
            self._selection, targets, Ellipsizer(target_data), claim, synchronous_client, self._can_receive)
        if claim:
            self._have_token = True
        if not self._can_receive:
            return
        if target_data and claim:
            targets = target_data.keys()
            text_targets = tuple(x for x in targets if x in TEXT_TARGETS)
            for text_target in text_targets:
                dtype, dformat, data = target_data.get(text_target)
                if dformat != 8:
                    continue
                text = str(data)
                if isinstance(data, bytes):
                    if text_target.lower().find("utf8") >= 0:
                        try:
                            text = data.decode("utf8")
                        except UnicodeDecodeError:
                            pass
                    else:
                        try:
                            text = data.decode("latin1")
                        except UnicodeDecodeError:
                            pass
                log("setting text data %s / %s of size %i: %s", dtype, dformat, len(text), Ellipsizer(text))
                self._owner_change_embargo = monotonic()
                self.clipboard.set_text(text, len(text))
                return
            # we should handle more datatypes here..

    ############################################################################
    # forward local requests to the remote clipboard:
    ############################################################################
    def schedule_emit_token(self, min_delay=0):
        def send_token(*token_data):
            self._have_token = False
            self.emit("send-clipboard-token", token_data)

        if not (self._want_targets or self._greedy_client):
            send_token()
            return
        # we need the targets:
        targets = self.clipboard.wait_for_targets()
        if not targets:
            send_token()
            return
        if not self._greedy_client:
            send_token(targets)
            return
        # for now, we only handle text targets:
        text_targets = tuple(x for x in targets if x in TEXT_TARGETS)
        if text_targets:
            text = self.clipboard.wait_for_text()
            if text:
                # should verify the target is actually utf8...
                text_target = text_targets[0]
                send_token(targets, (text_target, "UTF8_STRING", 8, text))
                return
        send_token(text_targets)

    def owner_change(self, clipboard, event):
        log("owner_change(%s, %s) window=%s, selection=%s",
            clipboard, event, event.window, event.selection)
        self.do_owner_changed()

    def do_owner_changed(self):
        elapsed = monotonic() - self._owner_change_embargo
        log("do_owner_changed() enabled=%s, elapsed=%s",
            self._enabled, elapsed)
        if not self._enabled or elapsed < BLOCK_DELAY:
            return
        self.schedule_emit_token()

    def get_contents(self, target: str, got_contents: ClipboardCallback, time=0):
        log("get_contents(%s, %s, %i) have-token=%s",
            target, got_contents, time, self._have_token)
        if target == "TARGETS":
            r = self.clipboard.wait_for_targets()
            if r and len(r) == 2 and r[0]:
                targets = r[1]
                atoms = tuple(x.name() for x in targets)
                got_contents("ATOM", 32, atoms)
                return
        elif target in TEXT_TARGETS:
            text = self.clipboard.wait_for_text()
            if text:
                got_contents(target, 8, text)
                return
        else:
            # data = wait_for_contents(target)?
            pass
        got_contents(target, 0, b"")


GObject.type_register(GTKClipboardProxy)



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

control = True
notifications = True
webcam = True
clipboard = True
audio = True
av_sync = True
fileprint = True
mmap = True
ssl = True
ssh = True
input_devices = True
commands = True
gstreamer = True
x11 = True
dbus = True
encoding = True
logging = True
network_state = True
shell = False
display = True
windows = True
cursors = True
rfb = True
http = True



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
from threading import Event
from typing import Any

from xpra.net.rfb.const import RFBEncoding
from xpra.net.rfb.encode import raw_encode, tight_encode, tight_png, rgb222_encode
from xpra.net.protocol.socket_handler import PACKET_JOIN_SIZE
from xpra.net.common import PacketElement
from xpra.util.objects import AtomicInteger
from xpra.util.str_fn import csv, strtobytes, memoryview_to_bytes
from xpra.log import Logger

log = Logger("rfb")

counter = AtomicInteger()


class RFBSource:
    __slots__ = (
        "protocol", "close_event",
        "ui_client", "counter", "share", "uuid", "lock", "keyboard_config",
        "encodings", "quality", "pixel_format"
    )

    def __init__(self, protocol, share=False):
        self.protocol = protocol
        self.close_event = Event()
        self.ui_client = True
        self.counter = 0
        self.share = share
        self.uuid = "RFB%5i" % counter.increase()
        self.lock = False
        self.keyboard_config = None
        self.encodings = [RFBEncoding.RAW]
        self.pixel_format = (32, 24, 0, 1, 255, 255, 255, 16, 8, 0)
        self.quality = 0

    def get_info(self) -> dict[str, Any]:
        return {
            "protocol": "rfb",
            "uuid": self.uuid,
            "share": self.share,
        }

    def set_encodings(self, encodings) -> None:
        known_encodings = []
        unknown_encodings = []
        for v in encodings:
            try:
                known_encodings.append(RFBEncoding(v))
            except ValueError:
                unknown_encodings.append(v)
        self.encodings = known_encodings
        log("RFB encodings: %s", csv(self.encodings))
        if unknown_encodings:
            log("RFB %i unknown encodings: %s", len(unknown_encodings), csv(unknown_encodings))

    def set_pixel_format(self, pixel_format) -> None:
        # bpp, depth, bigendian, truecolor, rmax, gmax, bmax, rshift, bshift, gshift
        self.pixel_format = tuple(pixel_format)
        bpp, depth, bigendian, truecolor, rmax, gmax, bmax, rshift, bshift, gshift = pixel_format
        log(" pixel depth %i, %i bits per pixel", depth, bpp)
        log(" bigendian=%s, truecolor=%s", bool(bigendian), bool(truecolor))
        if truecolor:
            log(" RGB max: %s, shift: %s", (rmax, gmax, bmax), (rshift, bshift, gshift))

    def get_window_info(self, _wids) -> dict:
        return {}

    def is_closed(self) -> bool:
        return self.close_event.is_set()

    def close(self) -> None:
        self.close_event.set()

    def ping(self) -> None:
        """ ignore as there are no equivalent messages in RFB """

    def keys_changed(self) -> None:
        """ not implemented yet """

    def set_default_keymap(self):
        log("set_default_keymap() keyboard_config=%s", self.keyboard_config)
        if self.keyboard_config:
            self.keyboard_config.set_default_keymap()
        return self.keyboard_config

    def set_keymap(self, _current_keyboard_config, keys_pressed, _force=False, _translate_only=False):
        kc = self.keyboard_config
        kc.keys_pressed = keys_pressed
        kc.set_keymap(True)
        kc.owner = self.uuid

    def send_server_event(self, _event_type: str, *_args: PacketElement) -> None:
        """ ignore as there are no equivalent messages in RFB """

    def send_cursor(self) -> None:
        """ not implemented yet """

    def update_mouse(self, *args) -> None:
        log("update_mouse%s", args)

    def damage(self, _wid, window, x, y, w, h, options=None) -> None:
        polling = options and options.get("polling", False)
        p = self.protocol
        if polling and p is None or p.queue_size() >= 2:
            # very basic RFB update rate control,
            # if there are packets waiting already
            # we'll just process the next polling update instead:
            return
        if self.is_closed():
            return
        encode = raw_encode
        kwargs = {}
        if self.pixel_format[:2] != (32, 24):
            if self.pixel_format[:3] == (8, 6, 0):
                # crappy initial format chosen by realvnc
                encode = rgb222_encode
            else:
                log("damage: unsupported client pixel format: %s", self.pixel_format)
                return
        elif RFBEncoding.TIGHT_PNG in self.encodings:
            encode = tight_png
        elif RFBEncoding.TIGHT in self.encodings:
            encode = tight_encode
            kwargs = {"quality": self.quality}
        # doesn't work
        # elif RFBEncoding.ZLIB in self.encodings:
        #    encode = zlib_encode
        packets = encode(window, x, y, w, h, **kwargs)
        if not packets:
            return
        self.send_many(*packets)

    def send_many(self, *packets):
        # merge small packets together:
        joined = []

        def send_joined():
            if joined:
                self.send(b"".join(memoryview_to_bytes(p) for p in joined))
                joined[:] = []

        for packet in packets:
            joined.append(packet)
            if sum(len(p) for p in joined) > PACKET_JOIN_SIZE:
                # too much, can't be joined
                joined.pop()
                send_joined()
                self.send(packet)
        send_joined()

    def send_clipboard(self, text) -> None:
        nocr = strtobytes(text.replace("\r", ""))
        msg = struct.pack(b"!BBBBI", 3, 0, 0, 0, len(nocr)) + nocr
        self.send(msg)

    def bell(self, *_args) -> None:
        msg = struct.pack(b"!B", 2)
        self.send(msg)

    def send(self, msg) -> None:
        p = self.protocol
        if p:
            p.send(msg)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# @PydevCodeAnalysisIgnore
# pylint: disable=no-member

import os
import sys
from typing import Any
from time import monotonic
from threading import RLock
from collections.abc import Sequence

from xpra.codecs.constants import TransientCodecException
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, print_nested_dict
from xpra.util.env import envint, envbool, first_time, numpy_import_context
from xpra.platform.paths import (
    get_default_conf_dirs, get_system_conf_dirs, get_user_conf_dirs,
    get_resources_dir, get_app_dir,
)
from xpra.os_util import WIN32
from xpra.util.system import is_WSL
from xpra.util.io import load_binary_file
from xpra.log import Logger, consume_verbose_argv

log = Logger("cuda")

if WIN32 and not os.environ.get("CUDA_PATH") and getattr(sys, "frozen", None) in ("windows_exe", "console_exe", True):
    os.environ["CUDA_PATH"] = get_app_dir()

if is_WSL() and not envbool("XPRA_PYCUDA_WSL", False):
    raise ImportError("refusing to import pycuda on WSL, use `XPRA_PYCUDA_WSL=1` to override")

with numpy_import_context("CUDA context import"):
    import pycuda
    log(f"loaded pycuda successfully: {pycuda}")
    from pycuda.driver import (
        get_version, get_driver_version, mem_get_info,
        init,
        Device, device_attribute, ctx_flags,
        module_from_buffer, LogicError,
    )

MIN_FREE_MEMORY = envint("XPRA_CUDA_MIN_FREE_MEMORY", 10)

# record when we get failures/success:
DEVICE_STATE: dict[int, bool] = {}


def record_device_failure(device_id: int) -> None:
    DEVICE_STATE[device_id] = False


def record_device_success(device_id: int) -> None:
    DEVICE_STATE[device_id] = True


def device_info(d) -> str:
    if not d:
        return "None"
    return f"{d.name()} @ {d.pci_bus_id()}"


def pci_bus_id(d) -> str:
    if not d:
        return "None"
    return d.pci_bus_id()


def device_name(d) -> str:
    if not d:
        return "None"
    return d.name()


def compute_capability(d) -> int:
    smmajor, smminor = d.compute_capability()
    return (smmajor << 4) + smminor


def get_pycuda_version() -> Sequence[int]:
    return pycuda.VERSION


def get_pycuda_info() -> dict[str, Any]:
    init_all_devices()
    i = {
        "version": {
            "": pycuda.VERSION,
            "text": pycuda.VERSION_TEXT,
        }
    }
    if pycuda.VERSION_STATUS:
        i["version.status"] = pycuda.VERSION_STATUS
    return i


def get_cuda_info() -> dict[str, Any]:
    init_all_devices()
    return {
        "driver": {
            "version": get_version(),
            "driver_version": get_driver_version(),
        }
    }


DEVICE_INFO: dict[int, str] = {}


def get_device_info(i: int) -> str:
    return DEVICE_INFO.get(i, "")


DEVICE_NAME: dict[int, str] = {}


def get_device_name(i: int) -> str:
    return DEVICE_NAME.get(i, "")


PREFS = None


def get_prefs() -> dict[str, Any]:
    global PREFS
    if PREFS is None:
        PREFS = do_get_prefs()
    return PREFS or {}


def do_get_prefs() -> dict[str, Any]:
    prefs: dict[str, Any] = {}
    dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs()
    log(f"get_prefs() will try to load cuda.conf from: {dirs}")
    for d in dirs:
        conf_file = os.path.join(os.path.expanduser(d), "cuda.conf")
        if not os.path.exists(conf_file):
            log(f"get_prefs() {conf_file!r} does not exist!")
            continue
        if not os.path.isfile(conf_file):
            log(f"get_prefs() {conf_file!r} is not a file!")
            continue
        c_prefs: dict[str, Any] = {}
        try:
            with open(conf_file, "rb") as f:
                for line in f:
                    sline = line.strip().rstrip(b'\r\n').strip().decode("latin1")
                    props = sline.split("=", 1)
                    if len(props) != 2:
                        continue
                    name = props[0].strip()
                    value = props[1].strip()
                    if name in ("enabled-devices", "disabled-devices"):
                        for v in value.split(","):
                            c_prefs.setdefault(name, []).append(v.strip())
                    elif name in ("device-id", "device-name", "load-balancing"):
                        c_prefs[name] = value
        except Exception as e:
            log.error(f"Error: cannot read cuda configuration file {conf_file!r}")
            log.estr(e)
        log(f"get_prefs() {conf_file!r} : {c_prefs}")
        prefs.update(c_prefs)
    return prefs


def get_pref(name: str):
    assert name in ("device-id", "device-name", "enabled-devices", "disabled-devices", "load-balancing")
    # ie: env_name("device-id")="XPRA_CUDA_DEVICE_ID"
    env_name = "XPRA_CUDA_" + str(name).upper().replace("-", "_")
    env_value = os.environ.get(env_name)
    if env_value is not None:
        if name in ("enabled-devices", "disabled-devices"):
            return env_value.split(",")
        return env_value
    return get_prefs().get(name)


def get_gpu_list(list_type):
    v = get_pref(list_type)
    log(f"get_gpu_list({list_type}) pref={v}")
    if not v:
        return None
    if "all" in v:
        return True
    if "none" in v:
        return []

    def dev(x):
        try:
            return int(x)
        except ValueError:
            return x.strip()

    try:
        return [dev(x) for x in v]
    except ValueError:
        log(f"get_gpu_list({list_type})", exc_info=True)
        log.error(f"Error: invalid value for {list_type!r} CUDA preference")
        return None


driver_init_done = None


def driver_init() -> bool:
    global driver_init_done
    if driver_init_done is None:
        log.info("CUDA initialization (this may take a few seconds)")
        try:
            init()
            driver_init_done = True
            log(f"CUDA driver version={get_driver_version()}")
            ngpus = Device.count()
            if ngpus == 0:
                cuda_v = ".".join(str(x) for x in get_version())
                log.info(f"CUDA {cuda_v} / PyCUDA {pycuda.VERSION_TEXT}, no devices found")
            driver_init_done = True
        except Exception as e:
            log("driver_init()", exc_info=True)
            log.warn("Warning: cannot initialize CUDA")
            log.warn(f" {e}")
            driver_init_done = False
    return driver_init_done


DEVICES: list[int] | None = None


def init_all_devices() -> list[int]:
    global DEVICES, DEVICE_INFO
    if DEVICES is not None:
        return DEVICES
    DEVICES = []
    DEVICE_INFO = {}
    enabled_gpus = get_gpu_list("enabled-devices")
    disabled_gpus = get_gpu_list("disabled-devices")
    if disabled_gpus is True or enabled_gpus == []:
        log("all devices are disabled!")
        return DEVICES
    log(f"init_all_devices() enabled: {csv(enabled_gpus)}, disabled: %s", csv(disabled_gpus) or "none")
    if not driver_init():
        return DEVICES
    ngpus = Device.count()
    log(f"init_all_devices() ngpus={ngpus}")
    if ngpus == 0:
        return DEVICES
    for i in range(ngpus):
        # shortcut if this GPU number is disabled:
        if disabled_gpus is not None and i in disabled_gpus:
            log(f"device {i} is in the list of disabled gpus, skipped")
            continue
        devinfo = f"gpu {i}"
        try:
            device = Device(i)
            devinfo = device_info(device)
            log(" + testing device %s: %s", i, devinfo)
            DEVICE_NAME[i] = device_name(device)
            DEVICE_INFO[i] = devinfo
            if check_device(i, device):
                DEVICES.append(i)
        except Exception as e:
            log.error("error on device %s: %s", devinfo, e)
    return DEVICES


def check_device(i: int, device, min_compute: int = 0) -> bool:
    ngpus = Device.count()
    da = device_attribute
    devinfo = device_info(device)
    devname = device_name(device)
    pci = pci_bus_id(device)
    host_mem = device.get_attribute(da.CAN_MAP_HOST_MEMORY)
    if not host_mem:
        log.warn("skipping device %s (cannot map host memory)", devinfo)
        return False
    compute = compute_capability(device)
    if compute < min_compute:
        log("ignoring device %s: compute capability %#x (minimum %#x required)",
            device_info(device), compute, min_compute)
        return False
    enabled_gpus = get_gpu_list("enabled-devices")
    disabled_gpus = get_gpu_list("disabled-devices")
    if enabled_gpus not in (None, True):
        # check the enabled gpu list:
        if not any(x in enabled_gpus for x in (i, devname, pci)):
            log("device %i '%s' / '%s' is not in the list of enabled gpus, skipped", i, devname, pci)
            return False
    if disabled_gpus is not None and (devname in disabled_gpus or pci in disabled_gpus):
        log("device '%s' / '%s' is in the list of disabled gpus, skipped", i, devname, pci)
        return False
    cf = ctx_flags
    context = device.make_context(flags=cf.SCHED_YIELD | cf.MAP_HOST)
    try:
        log("   created context=%s", context)
        log("   api version=%s", context.get_api_version())
        free, total = mem_get_info()
        log("   memory: free=%sMB, total=%sMB", int(free // 1024 // 1024), int(total // 1024 // 1024))
        log("   multi-processors: %s, clock rate: %s",
            device.get_attribute(da.MULTIPROCESSOR_COUNT), device.get_attribute(da.CLOCK_RATE))
        log("   max block sizes: (%s, %s, %s)",
            device.get_attribute(da.MAX_BLOCK_DIM_X),
            device.get_attribute(da.MAX_BLOCK_DIM_Y),
            device.get_attribute(da.MAX_BLOCK_DIM_Z),
            )
        log("   max grid sizes: (%s, %s, %s)",
            device.get_attribute(da.MAX_GRID_DIM_X),
            device.get_attribute(da.MAX_GRID_DIM_Y),
            device.get_attribute(da.MAX_GRID_DIM_Z),
            )
        max_width = device.get_attribute(da.MAXIMUM_TEXTURE2D_WIDTH)
        max_height = device.get_attribute(da.MAXIMUM_TEXTURE2D_HEIGHT)
        log("   maximum texture size: %sx%s", max_width, max_height)
        log("   max pitch: %s", device.get_attribute(da.MAX_PITCH))
        SMmajor, SMminor = device.compute_capability()
        compute = (SMmajor << 4) + SMminor
        log("   compute capability: %#x (%s.%s)", compute, SMmajor, SMminor)
        if i == 0:
            # we print the list info "header" from inside the loop
            # so that the log output is bunched up together
            log.info("CUDA %s / PyCUDA %s, found %s devices:",
                     ".".join([str(x) for x in get_version()]), pycuda.VERSION_TEXT, ngpus)
        mfree = round(100 * free / total)
        log.info(f"  + {device_info(device)} (memory: {mfree}% free, compute: {SMmajor}.{SMminor})")
        if SMmajor < 2:
            log.info("  this device is too old!")
            return False
        return True
    finally:
        context.pop()


def get_devices() -> list[int] | None:
    return DEVICES


def check_devices() -> None:
    devices = init_all_devices()
    assert devices, "no valid CUDA devices found!"


def reset_state() -> None:
    log("cuda_context.reset_state()")
    global DEVICE_STATE
    DEVICE_STATE = {}


def select_device(preferred_device_id=-1, min_compute=0) -> tuple[int, Any]:
    log("select_device(%s, %s)", preferred_device_id, min_compute)
    for device_id in (preferred_device_id, get_pref("device-id")):
        if device_id is not None and device_id >= 0:
            dct = make_device_context(device_id)
            if dct:
                device, context, tpct = dct
                context.pop()
                context.detach()
                if min_compute > 0:
                    compute = compute_capability(device)
                    if compute < min_compute:
                        log.warn("Warning: GPU device %i only supports compute %#x", device_id, compute)
                if tpct < MIN_FREE_MEMORY:
                    log.warn(f"Warning: GPU device {device_id} is low on memory: {tpct}%")
                return device_id, device
    load_balancing = get_pref("load-balancing")
    log("load-balancing=%s", load_balancing)
    if load_balancing == "round-robin":
        return select_round_robin(min_compute)
    if load_balancing and load_balancing != "memory" and first_time("cuda-load-balancing"):
        log.warn("Warning: invalid load balancing value '%s'", load_balancing)
    return select_best_free_memory(min_compute)


rr = 0


def select_round_robin(min_compute: int) -> tuple[int, Any]:
    if not driver_init():
        return -1, None
    enabled_gpus = get_gpu_list("enabled-devices")
    disabled_gpus = get_gpu_list("disabled-devices")
    if disabled_gpus is True or enabled_gpus == []:
        log("all devices are disabled!")
        return -1, None
    ngpus = Device.count()
    if ngpus == 0:
        return -1, None
    devices = list(range(ngpus))
    global rr
    i = rr
    device_id = 0
    device = None
    while devices:
        n = len(devices)
        i = (rr + 1) % n
        device_id = devices[i]
        device = Device(device_id)
        if check_device(device_id, device, min_compute):
            break
        devices.remove(device_id)
    rr = i
    return device_id, device


def select_best_free_memory(min_compute: int = 0) -> tuple[int, Any]:
    # load preferences:
    preferred_device_name = get_pref("device-name")
    devices = init_all_devices()
    free_pct = 0
    # split device list according to device state:
    ok_devices = [device_id for device_id in devices if DEVICE_STATE.get(device_id, True) is True]
    nok_devices = [device_id for device_id in devices if DEVICE_STATE.get(device_id, True) is not True]
    for list_name, device_list in {"OK": ok_devices, "failing": nok_devices}.items():
        selected_device_id = -1
        selected_device = None
        log("will test %s devices from %s list: %s", len(device_list), list_name, device_list)
        for device_id in device_list:
            context = None
            dct = make_device_context(device_id)
            if not dct:
                continue
            try:
                device, context, tpct = dct
                compute = compute_capability(device)
                if compute < min_compute:
                    log("ignoring device %s: compute capability %#x (minimum %#x required)",
                        device_info(device), compute, min_compute)
                elif preferred_device_name and device_info(device).find(preferred_device_name) >= 0:
                    log("device matches preferred device name: %s", preferred_device_name)
                    return device_id, device
                elif tpct >= MIN_FREE_MEMORY and tpct > free_pct:
                    log("device has enough free memory: %i (min=%i, current best device=%i)",
                        tpct, MIN_FREE_MEMORY, free_pct)
                    selected_device = device
                    selected_device_id = device_id
                    free_pct = tpct
            finally:
                if context:
                    context.pop()
                    context.detach()
        if selected_device_id >= 0 and selected_device:
            log_fn = log.info if len(devices) > 1 else log.debug
            log_fn("selected device %s: %s", selected_device_id, device_info(selected_device))
            return selected_device_id, selected_device
    return -1, None


def load_device(device_id: int):
    log("load_device(%i)", device_id)
    try:
        return Device(device_id)
    except Exception as e:
        log("load_device(%s)", device_id, exc_info=True)
        log.error("Error: allocating CUDA device %s", device_id)
        log.estr(e)
    return None


def make_device_context(device_id: int) -> tuple:
    log(f"make_device_context({device_id}")
    device = load_device(device_id)
    if not device:
        return ()
    log(f"make_device_context({device_id}) device_info={device_info(device)}")
    cf = ctx_flags
    flags = cf.SCHED_YIELD | cf.MAP_HOST
    try:
        context = device.make_context(flags=flags)
    except Exception as e:
        log(f"{device}.make_context({flags:x})", exc_info=True)
        log.error(f"Error: cannot create CUDA context for device {device_id}")
        log.estr(e)
        return ()
    log(f"created context={context}")
    free, total = mem_get_info()
    log("memory: free=%sMB, total=%sMB", int(free / 1024 / 1024), int(total / 1024 / 1024))
    tpct = 100 * free // total
    return device, context, tpct


def get_device_context(options: typedict):
    MIN_COMPUTE = 0x30
    device_id, device = select_device(options.intget("cuda_device", -1), min_compute=MIN_COMPUTE)
    if device_id < 0 or not device:
        return None
    return cuda_device_context(device_id, device)


default_device_context = None


def get_default_device_context():
    global default_device_context
    if default_device_context is None:
        start = monotonic()
        cuda_device_id, cuda_device = select_device()
        if cuda_device_id < 0 or not cuda_device:
            raise RuntimeError("failed to select a cuda device")
        log("using device %s", cuda_device)
        default_device_context = cuda_device_context(cuda_device_id, cuda_device)
        end = monotonic()
        log("default device context init took %.1fms", 1000 * (end - start))
    return default_device_context


def free_default_device_context() -> None:
    global default_device_context
    ddc = default_device_context
    default_device_context = None
    if ddc:
        ddc.free()


class cuda_device_context:
    __slots__ = ("device_id", "device", "context", "lock", "opengl")

    def __init__(self, device_id: int, device, opengl=False):
        assert device, "no cuda device"
        self.device_id = device_id
        self.device = device
        self.opengl = opengl
        self.context = None
        self.lock = RLock()
        log("%r", self)

    def __bool__(self):
        return self.device is not None

    def __enter__(self):
        if not self.lock.acquire(False):
            raise TransientCodecException("failed to acquire cuda device lock")
        if not self.context:
            self.make_context()
        return self.push_context()

    def make_context(self) -> None:
        start = monotonic()
        if self.opengl:
            with numpy_import_context("CUDA make context"):
                from pycuda import gl  # @UnresolvedImport pylint: disable=import-outside-toplevel
                self.context = gl.make_context(self.device)
        else:
            self.context = self.device.make_context(flags=ctx_flags.SCHED_YIELD | ctx_flags.MAP_HOST)
        end = monotonic()
        self.context.pop()
        log("cuda context allocation took %ims", 1000 * (end - start))

    def push_context(self):
        self.context.push()
        return self.context

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop_context()
        self.lock.release()

    def pop_context(self) -> None:
        c = self.context
        if c:
            c.pop()
        # except driver.LogicError as e:
        # log.warn("Warning: PyCUDA %s", e)
        # self.clean()
        # self.init_cuda()

    def __repr__(self):
        extra = " - locked" if self.lock._is_owned() else ""
        return f"cuda_device_context({self.device_id}{extra})"

    def get_info(self) -> dict[str, Any]:
        info: dict[str, Any] = {
            "id": self.device_id,
            "device": {
                "name": self.device.name(),
                "pci_bus_id": self.device.pci_bus_id(),
                "memory": int(self.device.total_memory() // 1024 // 1024),
            },
            "opengl": self.opengl,
        }
        if self.context:
            info["api_version"] = self.context.get_api_version()
        return info

    def __del__(self):
        self.free()

    def free(self) -> None:
        c = self.context
        if log:
            log("free() context=%s", c)
        if c:
            self.device_id = 0
            self.device = None
            self.context = None
            with self.lock:
                c.detach()


# cache kernel fatbin files:
KERNELS: dict[str, bytes] = {}


def get_CUDA_function(function_name: str):
    """
        Returns the compiled kernel for the given device
        and kernel key.
    """
    data = KERNELS.get(function_name)
    if data is None:
        cubin_file = os.path.join(get_resources_dir(), "cuda", f"{function_name}.fatbin")
        log(f"get_CUDA_function({function_name}) cubin file={cubin_file!r}")
        if not os.path.exists(cubin_file):
            log.error(f"Error: failed to load CUDA bin file {cubin_file!r}")
            log.error(" this file does not exist")
            return None
        data = load_binary_file(cubin_file)
        if not data:
            log.error(f"Error: failed to load CUDA bin file {cubin_file!r}")
            return None
        log(f" loaded {len(data)} bytes")
        KERNELS[function_name] = data
    # now load from cubin:
    start = monotonic()
    try:
        mod = module_from_buffer(data)
    except Exception as e:
        log(f"module_from_buffer({data})", exc_info=True)
        log.error(f"Error: failed to load module from buffer for {function_name!r}")
        log.estr(e)
        return None
    log(f"get_CUDA_function({function_name}) module={mod}")
    try:
        fn = function_name
        CUDA_function = mod.get_function(fn)
    except LogicError as e:
        raise ValueError(f"failed to load {function_name!r} from {mod}: {e}") from None
    end = monotonic()
    log(f"loading function {function_name!r} from pre-compiled cubin took %.1fms", 1000.0 * (end - start))
    return CUDA_function


def main() -> None:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("CUDA-Info", "CUDA Info"):
        consume_verbose_argv(sys.argv, "cuda")
        pycuda_info = get_pycuda_info()
        log.info("pycuda_info")
        print_nested_dict(pycuda_info, print_fn=log.info)
        log.info("cuda_info")
        print_nested_dict(get_cuda_info(), print_fn=log.info)
        log.info("preferences:")
        print_nested_dict(get_prefs(), print_fn=log.info)
        log.info("device automatically selected:")
        log.info(" %s", device_info(select_device()[1]))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from subprocess import PIPE, Popen


def get_status_output(*args, **kwargs):
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = PIPE
    try:
        p = Popen(*args, **kwargs)
    except Exception as e:
        print("error running %s,%s: %s" % (args, kwargs, e))
        return -1, "", ""
    stdout, stderr = p.communicate()
    return p.returncode, stdout, stderr


def get_output_lines(cmd: list[str]) -> list[str]:
    try:
        returncode, stdout, stderr = get_status_output(cmd)
        if returncode != 0:
            print("'%s' failed with return code %s" % (cmd, returncode))
            print("stderr: %s" % stderr)
        elif not stdout:
            print("could not get command output")
        else:
            out = stdout.decode('utf-8')
            return out.splitlines()
    except OSError:
        pass
    return []


def gen_range(minw: int, maxw: int, wstep: int, minh: int, maxh: int, hstep: int) -> None:
    max_ratio = 3
    max_hz = 160
    for w in range(minw, maxw, wstep):
        for h in range(minh, maxh, hstep):
            if float(w) / float(h) > max_ratio or float(h) / float(w) > max_ratio:
                continue
            pixel_clock = 300 * 1000 * 1000  #300MHz target
            #theoretical pixel_clock = w*h*hz
            #overhead for safety margins (timings):
            margin = 25
            hz = pixel_clock // (w * h * (100 + margin) // 100)
            hz = min(max_hz, hz)
            if hz > 100:
                hz = (hz // 10) * 10
            elif hz > 50:
                hz = (hz // 5) * 5
            print_cvt(w, h, hz)


def print_cvt(w: int, h: int, hz: int) -> None:
    cmd = ["cvt", str(w), str(h), str(hz)]
    lines = get_output_lines(cmd)
    for line in lines:
        if line.startswith("Modeline"):
            parts = line.split(" ")
            parts = [x for x in parts if len(x) > 0]
            parts[0] = "  Modeline"
            parts[1] = ('"%ix%i@%i"' % (w, h, hz)).ljust(15)
            parts[2] = parts[2].rjust(8)
            for i in range(3, len(parts)):
                parts[i] = parts[i].rjust(5)
            print(" ".join(parts))


def main() -> None:
    #lower resolutions (up to 4096x2048) with a 64 pixel step:
    gen_range(640, 4096, 64, 640, 2048, 64)
    #higher resolutions with 128 pixel step:
    gen_range(4096, 8192, 128, 2048, 4096, 128)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# ruff: noqa: F601

# found those here:
# http://wiki.winehq.org/list_Of_Windows_Messages
# which was extracted from MSDN somehow
WNDPROC_EVENT_NAMES = {
    0x0000: "WM_NULL",
    0x0001: "WM_CREATE",
    0x0002: "WM_DESTROY",
    0x0003: "WM_MOVE",
    0x0005: "WM_SIZE",
    0x0006: "WM_ACTIVATE",
    0x0007: "WM_SETFOCUS",
    0x0008: "WM_KILLFOCUS",
    0x000a: "WM_ENABLE",
    0x000b: "WM_SETREDRAW",
    0x000c: "WM_SETTEXT",
    0x000d: "WM_GETTEXT",
    0x000e: "WM_GETTEXTLENGTH",
    0x000f: "WM_PAINT",
    0x0010: "WM_CLOSE",
    0x0011: "WM_QUERYENDSESSION",
    0x0012: "WM_QUIT",
    0x0013: "WM_QUERYOPEN",
    0x0014: "WM_ERASEBKGND",
    0x0015: "WM_SYSCOLORCHANGE",
    0x0016: "WM_ENDSESSION",
    0x0018: "WM_SHOWWINDOW",
    0x0019: "WM_CTLCOLOR",
    0x001a: "WM_WININICHANGE",
    0x001b: "WM_DEVMODECHANGE",
    0x001c: "WM_ACTIVATEAPP",
    0x001d: "WM_FONTCHANGE",
    0x001e: "WM_TIMECHANGE",
    0x001f: "WM_CANCELMODE",
    0x0020: "WM_SETCURSOR",
    0x0021: "WM_MOUSEACTIVATE",
    0x0022: "WM_CHILDACTIVATE",
    0x0023: "WM_QUEUESYNC",
    0x0024: "WM_GETMINMAXINFO",
    0x0026: "WM_PAINTICON",
    0x0027: "WM_ICONERASEBKGND",
    0x0028: "WM_NEXTDLGCTL",
    0x002a: "WM_SPOOLERSTATUS",
    0x002b: "WM_DRAWITEM",
    0x002c: "WM_MEASUREITEM",
    0x002d: "WM_DELETEITEM",
    0x002e: "WM_VKEYTOITEM",
    0x002f: "WM_CHARTOITEM",
    0x0030: "WM_SETFONT",
    0x0031: "WM_GETFONT",
    0x0032: "WM_SETHOTKEY",
    0x0033: "WM_GETHOTKEY",
    0x0037: "WM_QUERYDRAGICON",
    0x0039: "WM_COMPAREITEM",
    0x003d: "WM_GETOBJECT",
    0x0041: "WM_COMPACTING",
    0x0044: "WM_COMMNOTIFY",
    0x0046: "WM_WINDOWPOSCHANGING",
    0x0047: "WM_WINDOWPOSCHANGED",
    0x0048: "WM_POWER",
    0x0049: "WM_COPYGLOBALDATA",
    0x004a: "WM_COPYDATA",
    0x004b: "WM_CANCELJOURNAL",
    0x004e: "WM_NOTIFY",
    0x0050: "WM_INPUTLANGCHANGEREQUEST",
    0x0051: "WM_INPUTLANGCHANGE",
    0x0052: "WM_TCARD",
    0x0053: "WM_HELP",
    0x0054: "WM_USERCHANGED",
    0x0055: "WM_NOTIFYFORMAT",
    0x007b: "WM_CONTEXTMENU",
    0x007c: "WM_STYLECHANGING",
    0x007d: "WM_STYLECHANGED",
    0x007e: "WM_DISPLAYCHANGE",
    0x007f: "WM_GETICON",
    0x0080: "WM_SETICON",
    0x0081: "WM_NCCREATE",
    0x0082: "WM_NCDESTROY",
    0x0083: "WM_NCCALCSIZE",
    0x0084: "WM_NCHITTEST",
    0x0085: "WM_NCPAINT",
    0x0086: "WM_NCACTIVATE",
    0x0087: "WM_GETDLGCODE",
    0x0088: "WM_SYNCPAINT",
    0x00a0: "WM_NCMOUSEMOVE",
    0x00a1: "WM_NCLBUTTONDOWN",
    0x00a2: "WM_NCLBUTTONUP",
    0x00a3: "WM_NCLBUTTONDBLCLK",
    0x00a4: "WM_NCRBUTTONDOWN",
    0x00a5: "WM_NCRBUTTONUP",
    0x00a6: "WM_NCRBUTTONDBLCLK",
    0x00a7: "WM_NCMBUTTONDOWN",
    0x00a8: "WM_NCMBUTTONUP",
    0x00a9: "WM_NCMBUTTONDBLCLK",
    0x00ab: "WM_NCXBUTTONDOWN",
    0x00ac: "WM_NCXBUTTONUP",
    0x00ad: "WM_NCXBUTTONDBLCLK",
    0x00ae: "WM_NCUAHDRAWCAPTION",
    0x00af: "WM_NCUAHDRAWFRAME",
    0x00e0: "SBM_SETPOS",
    0x00e1: "SBM_GETPOS",
    0x00e2: "SBM_SETRANGE",
    0x00e3: "SBM_GETRANGE",
    0x00e4: "SBM_ENABLE_ARROWS",
    0x00e6: "SBM_SETRANGEREDRAW",
    0x00e9: "SBM_SETSCROLLINFO",
    0x00ea: "SBM_GETSCROLLINFO",
    0x00eb: "SBM_GETSCROLLBARINFO",
    0x00ff: "WM_INPUT",
    0x0100: "WM_KEYDOWN",
    0x0100: "WM_KEYFIRST",
    0x0101: "WM_KEYUP",
    0x0102: "WM_CHAR",
    0x0103: "WM_DEADCHAR",
    0x0104: "WM_SYSKEYDOWN",
    0x0105: "WM_SYSKEYUP",
    0x0106: "WM_SYSCHAR",
    0x0107: "WM_SYSDEADCHAR",
    0x0108: "WM_KEYLAST",
    0x0109: "WM_WNT_CONVERTREQUESTEX",
    0x010a: "WM_CONVERTREQUEST",
    0x010b: "WM_CONVERTRESULT",
    0x010c: "WM_INTERIM",
    0x010d: "WM_IME_STARTCOMPOSITION",
    0x010e: "WM_IME_ENDCOMPOSITION",
    0x010f: "WM_IME_COMPOSITION",
    0x010f: "WM_IME_KEYLAST",
    0x0110: "WM_INITDIALOG",
    0x0111: "WM_COMMAND",
    0x0112: "WM_SYSCOMMAND",
    0x0113: "WM_TIMER",
    0x0114: "WM_HSCROLL",
    0x0115: "WM_VSCROLL",
    0x0116: "WM_INITMENU",
    0x0117: "WM_INITMENUPOPUP",
    0x0118: "WM_SYSTIMER",
    0x011f: "WM_MENUSELECT",
    0x0120: "WM_MENUCHAR",
    0x0121: "WM_ENTERIDLE",
    0x0122: "WM_MENURBUTTONUP",
    0x0123: "WM_MENUDRAG",
    0x0124: "WM_MENUGETOBJECT",
    0x0125: "WM_UNINITMENUPOPUP",
    0x0126: "WM_MENUCOMMAND",
    0x0127: "WM_CHANGEUISTATE",
    0x0128: "WM_UPDATEUISTATE",
    0x0129: "WM_QUERYUISTATE",
    0x0132: "WM_CTLCOLORMSGBOX",
    0x0133: "WM_CTLCOLOREDIT",
    0x0134: "WM_CTLCOLORLISTBOX",
    0x0135: "WM_CTLCOLORBTN",
    0x0136: "WM_CTLCOLORDLG",
    0x0137: "WM_CTLCOLORSCROLLBAR",
    0x0138: "WM_CTLCOLORSTATIC",
    0x0200: "WM_MOUSEFIRST",
    0x0200: "WM_MOUSEMOVE",
    0x0201: "WM_LBUTTONDOWN",
    0x0202: "WM_LBUTTONUP",
    0x0203: "WM_LBUTTONDBLCLK",
    0x0204: "WM_RBUTTONDOWN",
    0x0205: "WM_RBUTTONUP",
    0x0206: "WM_RBUTTONDBLCLK",
    0x0207: "WM_MBUTTONDOWN",
    0x0208: "WM_MBUTTONUP",
    0x0209: "WM_MBUTTONDBLCLK",
    0x0209: "WM_MOUSELAST",
    0x020a: "WM_MOUSEWHEEL",
    0x020b: "WM_XBUTTONDOWN",
    0x020c: "WM_XBUTTONUP",
    0x020d: "WM_XBUTTONDBLCLK",
    0x0210: "WM_PARENTNOTIFY",
    0x0211: "WM_ENTERMENULOOP",
    0x0212: "WM_EXITMENULOOP",
    0x0213: "WM_NEXTMENU",
    0x0214: "WM_SIZING",
    0x0215: "WM_CAPTURECHANGED",
    0x0216: "WM_MOVING",
    0x0218: "WM_POWERBROADCAST",
    0x0219: "WM_DEVICECHANGE",
    0x0220: "WM_MDICREATE",
    0x0221: "WM_MDIDESTROY",
    0x0222: "WM_MDIACTIVATE",
    0x0223: "WM_MDIRESTORE",
    0x0224: "WM_MDINEXT",
    0x0225: "WM_MDIMAXIMIZE",
    0x0226: "WM_MDITILE",
    0x0227: "WM_MDICASCADE",
    0x0228: "WM_MDIICONARRANGE",
    0x0229: "WM_MDIGETACTIVE",
    0x0230: "WM_MDISETMENU",
    0x0231: "WM_ENTERSIZEMOVE",
    0x0232: "WM_EXITSIZEMOVE",
    0x0233: "WM_DROPFILES",
    0x0234: "WM_MDIREFRESHMENU",
    0x0280: "WM_IME_REPORT",
    0x0281: "WM_IME_SETCONTEXT",
    0x0282: "WM_IME_NOTIFY",
    0x0283: "WM_IME_CONTROL",
    0x0284: "WM_IME_COMPOSITIONFULL",
    0x0285: "WM_IME_SELECT",
    0x0286: "WM_IME_CHAR",
    0x0288: "WM_IME_REQUEST",
    0x0290: "WM_IMEKEYDOWN",
    0x0290: "WM_IME_KEYDOWN",
    0x0291: "WM_IMEKEYUP",
    0x0291: "WM_IME_KEYUP",
    0x02a0: "WM_NCMOUSEHOVER",
    0x02a1: "WM_MOUSEHOVER",
    0x02a2: "WM_NCMOUSELEAVE",
    0x02a3: "WM_MOUSELEAVE",
    0x02b1: "WM_WTSSESSION_CHANGE",
    0x0300: "WM_CUT",
    0x0301: "WM_COPY",
    0x0302: "WM_PASTE",
    0x0303: "WM_CLEAR",
    0x0304: "WM_UNDO",
    0x0305: "WM_RENDERFORMAT",
    0x0306: "WM_RENDERALLFORMATS",
    0x0307: "WM_DESTROYCLIPBOARD",
    0x0308: "WM_DRAWCLIPBOARD",
    0x0309: "WM_PAINTCLIPBOARD",
    0x030a: "WM_VSCROLLCLIPBOARD",
    0x030b: "WM_SIZECLIPBOARD",
    0x030c: "WM_ASKCBFORMATNAME",
    0x030d: "WM_CHANGECBCHAIN",
    0x030e: "WM_HSCROLLCLIPBOARD",
    0x030f: "WM_QUERYNEWPALETTE",
    0x0310: "WM_PALETTEISCHANGING",
    0x0311: "WM_PALETTECHANGED",
    0x0312: "WM_HOTKEY",
    0x0317: "WM_PRINT",
    0x0318: "WM_PRINTCLIENT",
    0x0319: "WM_APPCOMMAND",
    0x031E: "WM_DWMCOMPOSITIONCHANGED",
    0x0358: "WM_HANDHELDFIRST",
    0x035f: "WM_HANDHELDLAST",
    0x0360: "WM_AFXFIRST",
    0x037f: "WM_AFXLAST",
    0x0380: "WM_PENWINFIRST",
    0x0381: "WM_RCRESULT",
    0x0382: "WM_HOOKRCRESULT",
    0x0383: "WM_GLOBALRCCHANGE",
    0x0383: "WM_PENMISCINFO",
    0x0384: "WM_SKB",
    0x0385: "WM_HEDITCTL",
    0x0385: "WM_PENCTL",
    0x0386: "WM_PENMISC",
    0x0387: "WM_CTLINIT",
    0x0388: "WM_PENEVENT",
    0x038f: "WM_PENWINLAST",
    0x0400: "DDM_SETFMT",
    0x0400: "DM_GETDEFID",
    0x0400: "NIN_SELECT",
    0x0400: "TBM_GETPOS",
    0x0400: "WM_PSD_PAGESETUPDLG",
    0x0400: "WM_USER",
    0x0401: "CBEM_INSERTITEMA",
    0x0401: "DDM_DRAW",
    0x0401: "DM_SETDEFID",
    0x0401: "HKM_SETHOTKEY",
    0x0401: "PBM_SETRANGE",
    0x0401: "RB_INSERTBANDA",
    0x0401: "SB_SETTEXTA",
    0x0401: "TB_ENABLEBUTTON",
    0x0401: "TBM_GETRANGEMIN",
    0x0401: "TTM_ACTIVATE",
    0x0401: "WM_CHOOSEFONT_GETLOGFONT",
    0x0401: "WM_PSD_FULLPAGERECT",
    0x0402: "CBEM_SETIMAGELIST",
    0x0402: "DDM_CLOSE",
    0x0402: "DM_REPOSITION",
    0x0402: "HKM_GETHOTKEY",
    0x0402: "PBM_SETPOS",
    0x0402: "RB_DELETEBAND",
    0x0402: "SB_GETTEXTA",
    0x0402: "TB_CHECKBUTTON",
    0x0402: "TBM_GETRANGEMAX",
    0x0402: "WM_PSD_MINMARGINRECT",
    0x0403: "CBEM_GETIMAGELIST",
    0x0403: "DDM_BEGIN",
    0x0403: "HKM_SETRULES",
    0x0403: "PBM_DELTAPOS",
    0x0403: "RB_GETBARINFO",
    0x0403: "SB_GETTEXTLENGTHA",
    0x0403: "TBM_GETTIC",
    0x0403: "TB_PRESSBUTTON",
    0x0403: "TTM_SETDELAYTIME",
    0x0403: "WM_PSD_MARGINRECT",
    0x0404: "CBEM_GETITEMA",
    0x0404: "DDM_END",
    0x0404: "PBM_SETSTEP",
    0x0404: "RB_SETBARINFO",
    0x0404: "SB_SETPARTS",
    0x0404: "TB_HIDEBUTTON",
    0x0404: "TBM_SETTIC",
    0x0404: "TTM_ADDTOOLA",
    0x0404: "WM_PSD_GREEKTEXTRECT",
    0x0405: "CBEM_SETITEMA",
    0x0405: "PBM_STEPIT",
    0x0405: "TB_INDETERMINATE",
    0x0405: "TBM_SETPOS",
    0x0405: "TTM_DELTOOLA",
    0x0405: "WM_PSD_ENVSTAMPRECT",
    0x0406: "CBEM_GETCOMBOCONTROL",
    0x0406: "PBM_SETRANGE32",
    0x0406: "RB_SETBANDINFOA",
    0x0406: "SB_GETPARTS",
    0x0406: "TB_MARKBUTTON",
    0x0406: "TBM_SETRANGE",
    0x0406: "TTM_NEWTOOLRECTA",
    0x0406: "WM_PSD_YAFULLPAGERECT",
    0x0407: "CBEM_GETEDITCONTROL",
    0x0407: "PBM_GETRANGE",
    0x0407: "RB_SETPARENT",
    0x0407: "SB_GETBORDERS",
    0x0407: "TBM_SETRANGEMIN",
    0x0407: "TTM_RELAYEVENT",
    0x0408: "CBEM_SETEXSTYLE",
    0x0408: "PBM_GETPOS",
    0x0408: "RB_HITTEST",
    0x0408: "SB_SETMINHEIGHT",
    0x0408: "TBM_SETRANGEMAX",
    0x0408: "TTM_GETTOOLINFOA",
    0x0409: "CBEM_GETEXSTYLE",
    0x0409: "CBEM_GETEXTENDEDSTYLE",
    0x0409: "PBM_SETBARCOLOR",
    0x0409: "RB_GETRECT",
    0x0409: "SB_SIMPLE",
    0x0409: "TB_ISBUTTONENABLED",
    0x0409: "TBM_CLEARTICS",
    0x0409: "TTM_SETTOOLINFOA",
    0x040a: "CBEM_HASEDITCHANGED",
    0x040a: "RB_INSERTBANDW",
    0x040a: "SB_GETRECT",
    0x040a: "TB_ISBUTTONCHECKED",
    0x040a: "TBM_SETSEL",
    0x040a: "TTM_HITTESTA",
    0x040a: "WIZ_QUERYNUMPAGES",
    0x040b: "CBEM_INSERTITEMW",
    0x040b: "RB_SETBANDINFOW",
    0x040b: "SB_SETTEXTW",
    0x040b: "TB_ISBUTTONPRESSED",
    0x040b: "TBM_SETSELSTART",
    0x040b: "TTM_GETTEXTA",
    0x040b: "WIZ_NEXT",
    0x040c: "CBEM_SETITEMW",
    0x040c: "RB_GETBANDCOUNT",
    0x040c: "SB_GETTEXTLENGTHW",
    0x040c: "TB_ISBUTTONHIDDEN",
    0x040c: "TBM_SETSELEND",
    0x040c: "TTM_UPDATETIPTEXTA",
    0x040c: "WIZ_PREV",
    0x040d: "CBEM_GETITEMW",
    0x040d: "RB_GETROWCOUNT",
    0x040d: "SB_GETTEXTW",
    0x040d: "TB_ISBUTTONINDETERMINATE",
    0x040d: "TTM_GETTOOLCOUNT",
    0x040e: "CBEM_SETEXTENDEDSTYLE",
    0x040e: "RB_GETROWHEIGHT",
    0x040e: "SB_ISSIMPLE",
    0x040e: "TB_ISBUTTONHIGHLIGHTED",
    0x040e: "TBM_GETPTICS",
    0x040e: "TTM_ENUMTOOLSA",
    0x040f: "SB_SETICON",
    0x040f: "TBM_GETTICPOS",
    0x040f: "TTM_GETCURRENTTOOLA",
    0x0410: "RB_IDTOINDEX",
    0x0410: "SB_SETTIPTEXTA",
    0x0410: "TBM_GETNUMTICS",
    0x0410: "TTM_WINDOWFROMPOINT",
    0x0411: "RB_GETTOOLTIPS",
    0x0411: "SB_SETTIPTEXTW",
    0x0411: "TBM_GETSELSTART",
    0x0411: "TB_SETSTATE",
    0x0411: "TTM_TRACKACTIVATE",
    0x0412: "RB_SETTOOLTIPS",
    0x0412: "SB_GETTIPTEXTA",
    0x0412: "TB_GETSTATE",
    0x0412: "TBM_GETSELEND",
    0x0412: "TTM_TRACKPOSITION",
    0x0413: "RB_SETBKCOLOR",
    0x0413: "SB_GETTIPTEXTW",
    0x0413: "TB_ADDBITMAP",
    0x0413: "TBM_CLEARSEL",
    0x0413: "TTM_SETTIPBKCOLOR",
    0x0414: "RB_GETBKCOLOR",
    0x0414: "SB_GETICON",
    0x0414: "TB_ADDBUTTONSA",
    0x0414: "TBM_SETTICFREQ",
    0x0414: "TTM_SETTIPTEXTCOLOR",
    0x0415: "RB_SETTEXTCOLOR",
    0x0415: "TB_INSERTBUTTONA",
    0x0415: "TBM_SETPAGESIZE",
    0x0415: "TTM_GETDELAYTIME",
    0x0416: "RB_GETTEXTCOLOR",
    0x0416: "TB_DELETEBUTTON",
    0x0416: "TBM_GETPAGESIZE",
    0x0416: "TTM_GETTIPBKCOLOR",
    0x0417: "RB_SIZETORECT",
    0x0417: "TB_GETBUTTON",
    0x0417: "TBM_SETLINESIZE",
    0x0417: "TTM_GETTIPTEXTCOLOR",
    0x0418: "RB_BEGINDRAG",
    0x0418: "TB_BUTTONCOUNT",
    0x0418: "TBM_GETLINESIZE",
    0x0418: "TTM_SETMAXTIPWIDTH",
    0x0419: "RB_ENDDRAG",
    0x0419: "TB_COMMANDTOINDEX",
    0x0419: "TBM_GETTHUMBRECT",
    0x0419: "TTM_GETMAXTIPWIDTH",
    0x041a: "RB_DRAGMOVE",
    0x041a: "TBM_GETCHANNELRECT",
    0x041a: "TB_SAVERESTOREA",
    0x041a: "TTM_SETMARGIN",
    0x041b: "RB_GETBARHEIGHT",
    0x041b: "TB_CUSTOMIZE",
    0x041b: "TBM_SETTHUMBLENGTH",
    0x041b: "TTM_GETMARGIN",
    0x041c: "RB_GETBANDINFOW",
    0x041c: "TB_ADDSTRINGA",
    0x041c: "TBM_GETTHUMBLENGTH",
    0x041c: "TTM_POP",
    0x041d: "RB_GETBANDINFOA",
    0x041d: "TB_GETITEMRECT",
    0x041d: "TBM_SETTOOLTIPS",
    0x041d: "TTM_UPDATE",
    0x041e: "RB_MINIMIZEBAND",
    0x041e: "TB_BUTTONSTRUCTSIZE",
    0x041e: "TBM_GETTOOLTIPS",
    0x041e: "TTM_GETBUBBLESIZE",
    0x041f: "RB_MAXIMIZEBAND",
    0x041f: "TBM_SETTIPSIDE",
    0x041f: "TB_SETBUTTONSIZE",
    0x041f: "TTM_ADJUSTRECT",
    0x0420: "TBM_SETBUDDY",
    0x0420: "TB_SETBITMAPSIZE",
    0x0420: "TTM_SETTITLEA",
    0x0421: "MSG_FTS_JUMP_VA",
    0x0421: "TB_AUTOSIZE",
    0x0421: "TBM_GETBUDDY",
    0x0421: "TTM_SETTITLEW",
    0x0422: "RB_GETBANDBORDERS",
    0x0423: "MSG_FTS_JUMP_QWORD",
    0x0423: "RB_SHOWBAND",
    0x0423: "TB_GETTOOLTIPS",
    0x0424: "MSG_REINDEX_REQUEST",
    0x0424: "TB_SETTOOLTIPS",
    0x0425: "MSG_FTS_WHERE_IS_IT",
    0x0425: "RB_SETPALETTE",
    0x0425: "TB_SETPARENT",
    0x0426: "RB_GETPALETTE",
    0x0427: "RB_MOVEBAND",
    0x0427: "TB_SETROWS",
    0x0428: "TB_GETROWS",
    0x0429: "TB_GETBITMAPFLAGS",
    0x042a: "TB_SETCMDID",
    0x042b: "RB_PUSHCHEVRON",
    0x042b: "TB_CHANGEBITMAP",
    0x042c: "TB_GETBITMAP",
    0x042d: "MSG_GET_DEFFONT",
    0x042d: "TB_GETBUTTONTEXTA",
    0x042e: "TB_REPLACEBITMAP",
    0x042f: "TB_SETINDENT",
    0x0430: "TB_SETIMAGELIST",
    0x0431: "TB_GETIMAGELIST",
    0x0432: "TB_LOADIMAGES",
    0x0432: "TTM_ADDTOOLW",
    0x0433: "TB_GETRECT",
    0x0433: "TTM_DELTOOLW",
    0x0434: "TB_SETHOTIMAGELIST",
    0x0434: "TTM_NEWTOOLRECTW",
    0x0435: "TB_GETHOTIMAGELIST",
    0x0435: "TTM_GETTOOLINFOW",
    0x0436: "TB_SETDISABLEDIMAGELIST",
    0x0436: "TTM_SETTOOLINFOW",
    0x0437: "TB_GETDISABLEDIMAGELIST",
    0x0437: "TTM_HITTESTW",
    0x0438: "TB_SETSTYLE",
    0x0438: "TTM_GETTEXTW",
    0x0439: "TB_GETSTYLE",
    0x0439: "TTM_UPDATETIPTEXTW",
    0x043a: "TB_GETBUTTONSIZE",
    0x043a: "TTM_ENUMTOOLSW",
    0x043b: "TB_SETBUTTONWIDTH",
    0x043b: "TTM_GETCURRENTTOOLW",
    0x043c: "TB_SETMAXTEXTROWS",
    0x043d: "TB_GETTEXTROWS",
    0x043e: "TB_GETOBJECT",
    0x043f: "TB_GETBUTTONINFOW",
    0x0440: "TB_SETBUTTONINFOW",
    0x0441: "TB_GETBUTTONINFOA",
    0x0442: "TB_SETBUTTONINFOA",
    0x0443: "TB_INSERTBUTTONW",
    0x0444: "TB_ADDBUTTONSW",
    0x0445: "TB_HITTEST",
    0x0446: "TB_SETDRAWTEXTFLAGS",
    0x0447: "TB_GETHOTITEM",
    0x0448: "TB_SETHOTITEM",
    0x0449: "TB_SETANCHORHIGHLIGHT",
    0x044a: "TB_GETANCHORHIGHLIGHT",
    0x044b: "TB_GETBUTTONTEXTW",
    0x044c: "TB_SAVERESTOREW",
    0x044d: "TB_ADDSTRINGW",
    0x044e: "TB_MAPACCELERATORA",
    0x044f: "TB_GETINSERTMARK",
    0x0450: "TB_SETINSERTMARK",
    0x0451: "TB_INSERTMARKHITTEST",
    0x0452: "TB_MOVEBUTTON",
    0x0453: "TB_GETMAXSIZE",
    0x0454: "TB_SETEXTENDEDSTYLE",
    0x0455: "TB_GETEXTENDEDSTYLE",
    0x0456: "TB_GETPADDING",
    0x0457: "TB_SETPADDING",
    0x0458: "TB_SETINSERTMARKCOLOR",
    0x0459: "TB_GETINSERTMARKCOLOR",
    0x045a: "TB_MAPACCELERATORW",
    0x045b: "TB_GETSTRINGW",
    0x045c: "TB_GETSTRINGA",
    0x0463: "TAPI_REPLY",
    0x0464: "ACM_OPENA",
    0x0464: "BFFM_SETSTATUSTEXTA",
    0x0464: "CDM_FIRST",
    0x0464: "CDM_GETSPEC",
    0x0464: "IPM_CLEARADDRESS",
    0x0464: "WM_CAP_UNICODE_START",
    0x0465: "ACM_PLAY",
    0x0465: "BFFM_ENABLEOK",
    0x0465: "CDM_GETFILEPATH",
    0x0465: "IPM_SETADDRESS",
    0x0465: "PSM_SETCURSEL",
    0x0465: "UDM_SETRANGE",
    0x0465: "WM_CHOOSEFONT_SETLOGFONT",
    0x0466: "ACM_STOP",
    0x0466: "BFFM_SETSELECTIONA",
    0x0466: "CDM_GETFOLDERPATH",
    0x0466: "IPM_GETADDRESS",
    0x0466: "PSM_REMOVEPAGE",
    0x0466: "UDM_GETRANGE",
    0x0466: "WM_CAP_SET_CALLBACK_ERRORW",
    0x0466: "WM_CHOOSEFONT_SETFLAGS",
    0x0467: "ACM_OPENW",
    0x0467: "BFFM_SETSELECTIONW",
    0x0467: "CDM_GETFOLDERIDLIST",
    0x0467: "IPM_SETRANGE",
    0x0467: "PSM_ADDPAGE",
    0x0467: "UDM_SETPOS",
    0x0467: "WM_CAP_SET_CALLBACK_STATUSW",
    0x0468: "BFFM_SETSTATUSTEXTW",
    0x0468: "CDM_SETCONTROLTEXT",
    0x0468: "IPM_SETFOCUS",
    0x0468: "PSM_CHANGED",
    0x0468: "UDM_GETPOS",
    0x0469: "CDM_HIDECONTROL",
    0x0469: "IPM_ISBLANK",
    0x0469: "PSM_RESTARTWINDOWS",
    0x0469: "UDM_SETBUDDY",
    0x046a: "CDM_SETDEFEXT",
    0x046a: "PSM_REBOOTSYSTEM",
    0x046a: "UDM_GETBUDDY",
    0x046b: "PSM_CANCELTOCLOSE",
    0x046b: "UDM_SETACCEL",
    0x046c: "EM_CONVPOSITION",
    0x046c: "PSM_QUERYSIBLINGS",
    0x046c: "UDM_GETACCEL",
    0x046d: "MCIWNDM_GETZOOM",
    0x046d: "PSM_UNCHANGED",
    0x046d: "UDM_SETBASE",
    0x046e: "PSM_APPLY",
    0x046e: "UDM_GETBASE",
    0x046f: "PSM_SETTITLEA",
    0x046f: "UDM_SETRANGE32",
    0x0470: "PSM_SETWIZBUTTONS",
    0x0470: "UDM_GETRANGE32",
    0x0470: "WM_CAP_DRIVER_GET_NAMEW",
    0x0471: "PSM_PRESSBUTTON",
    0x0471: "UDM_SETPOS32",
    0x0471: "WM_CAP_DRIVER_GET_VERSIONW",
    0x0472: "PSM_SETCURSELID",
    0x0472: "UDM_GETPOS32",
    0x0473: "PSM_SETFINISHTEXTA",
    0x0474: "PSM_GETTABCONTROL",
    0x0475: "PSM_ISDIALOGMESSAGE",
    0x0476: "MCIWNDM_REALIZE",
    0x0476: "PSM_GETCURRENTPAGEHWND",
    0x0477: "MCIWNDM_SETTIMEFORMATA",
    0x0477: "PSM_INSERTPAGE",
    0x0478: "MCIWNDM_GETTIMEFORMATA",
    0x0478: "PSM_SETTITLEW",
    0x0478: "WM_CAP_FILE_SET_CAPTURE_FILEW",
    0x0479: "MCIWNDM_VALIDATEMEDIA",
    0x0479: "PSM_SETFINISHTEXTW",
    0x0479: "WM_CAP_FILE_GET_CAPTURE_FILEW",
    0x047b: "MCIWNDM_PLAYTO",
    0x047b: "WM_CAP_FILE_SAVEASW",
    0x047c: "MCIWNDM_GETFILENAMEA",
    0x047d: "MCIWNDM_GETDEVICEA",
    0x047d: "PSM_SETHEADERTITLEA",
    0x047d: "WM_CAP_FILE_SAVEDIBW",
    0x047e: "MCIWNDM_GETPALETTE",
    0x047e: "PSM_SETHEADERTITLEW",
    0x047f: "MCIWNDM_SETPALETTE",
    0x047f: "PSM_SETHEADERSUBTITLEA",
    0x0480: "MCIWNDM_GETERRORA",
    0x0480: "PSM_SETHEADERSUBTITLEW",
    0x0481: "PSM_HWNDTOINDEX",
    0x0482: "PSM_INDEXTOHWND",
    0x0483: "MCIWNDM_SETINACTIVETIMER",
    0x0483: "PSM_PAGETOINDEX",
    0x0484: "PSM_INDEXTOPAGE",
    0x0485: "DL_BEGINDRAG",
    0x0485: "MCIWNDM_GETINACTIVETIMER",
    0x0485: "PSM_IDTOINDEX",
    0x0486: "DL_DRAGGING",
    0x0486: "PSM_INDEXTOID",
    0x0487: "DL_DROPPED",
    0x0487: "PSM_GETRESULT",
    0x0488: "DL_CANCELDRAG",
    0x0488: "PSM_RECALCPAGESIZES",
    0x048c: "MCIWNDM_GET_SOURCE",
    0x048d: "MCIWNDM_PUT_SOURCE",
    0x048e: "MCIWNDM_GET_DEST",
    0x048f: "MCIWNDM_PUT_DEST",
    0x0490: "MCIWNDM_CAN_PLAY",
    0x0491: "MCIWNDM_CAN_WINDOW",
    0x0492: "MCIWNDM_CAN_RECORD",
    0x0493: "MCIWNDM_CAN_SAVE",
    0x0494: "MCIWNDM_CAN_EJECT",
    0x0495: "MCIWNDM_CAN_CONFIG",
    0x0496: "IE_GETINK",
    0x0496: "IE_MSGFIRST",
    0x0496: "MCIWNDM_PALETTEKICK",
    0x0497: "IE_SETINK",
    0x0498: "IE_GETPENTIP",
    0x0499: "IE_SETPENTIP",
    0x049a: "IE_GETERASERTIP",
    0x049b: "IE_SETERASERTIP",
    0x049c: "IE_GETBKGND",
    0x049d: "IE_SETBKGND",
    0x049e: "IE_GETGRIDORIGIN",
    0x049f: "IE_SETGRIDORIGIN",
    0x04a0: "IE_GETGRIDPEN",
    0x04a1: "IE_SETGRIDPEN",
    0x04a2: "IE_GETGRIDSIZE",
    0x04a3: "IE_SETGRIDSIZE",
    0x04a4: "IE_GETMODE",
    0x04a5: "IE_SETMODE",
    0x04a6: "IE_GETINKRECT",
    0x04a6: "WM_CAP_SET_MCI_DEVICEW",
    0x04a7: "WM_CAP_GET_MCI_DEVICEW",
    0x04b4: "WM_CAP_PAL_OPENW",
    0x04b5: "WM_CAP_PAL_SAVEW",
    0x04b8: "IE_GETAPPDATA",
    0x04b9: "IE_SETAPPDATA",
    0x04ba: "IE_GETDRAWOPTS",
    0x04bb: "IE_SETDRAWOPTS",
    0x04bc: "IE_GETFORMAT",
    0x04bd: "IE_SETFORMAT",
    0x04be: "IE_GETINKINPUT",
    0x04bf: "IE_SETINKINPUT",
    0x04c0: "IE_GETNOTIFY",
    0x04c1: "IE_SETNOTIFY",
    0x04c2: "IE_GETRECOG",
    0x04c3: "IE_SETRECOG",
    0x04c4: "IE_GETSECURITY",
    0x04c5: "IE_SETSECURITY",
    0x04c6: "IE_GETSEL",
    0x04c7: "IE_SETSEL",
    0x04c8: "CDM_LAST",
    0x04c8: "IE_DOCOMMAND",
    0x04c8: "MCIWNDM_NOTIFYMODE",
    0x04c9: "IE_GETCOMMAND",
    0x04ca: "IE_GETCOUNT",
    0x04cb: "IE_GETGESTURE",
    0x04cb: "MCIWNDM_NOTIFYMEDIA",
    0x04cc: "IE_GETMENU",
    0x04cd: "IE_GETPAINTDC",
    0x04cd: "MCIWNDM_NOTIFYERROR",
    0x04ce: "IE_GETPDEVENT",
    0x04cf: "IE_GETSELCOUNT",
    0x04d0: "IE_GETSELITEMS",
    0x04d1: "IE_GETSTYLE",
    0x04db: "MCIWNDM_SETTIMEFORMATW",
    0x04dc: "EM_OUTLINE",
    0x04dc: "MCIWNDM_GETTIMEFORMATW",
    0x04dd: "EM_GETSCROLLPOS",
    0x04de: "EM_SETSCROLLPOS",
    0x04df: "EM_SETFONTSIZE",
    0x04e0: "MCIWNDM_GETFILENAMEW",
    0x04e1: "MCIWNDM_GETDEVICEW",
    0x04e4: "MCIWNDM_GETERRORW",
    0x0600: "FM_GETFOCUS",
    0x0601: "FM_GETDRIVEINFOA",
    0x0602: "FM_GETSELCOUNT",
    0x0603: "FM_GETSELCOUNTLFN",
    0x0604: "FM_GETFILESELA",
    0x0605: "FM_GETFILESELLFNA",
    0x0606: "FM_REFRESH_WINDOWS",
    0x0607: "FM_RELOAD_EXTENSIONS",
    0x0611: "FM_GETDRIVEINFOW",
    0x0614: "FM_GETFILESELW",
    0x0615: "FM_GETFILESELLFNW",
    0x0659: "WLX_WM_SAS",
    0x07e8: "SM_GETSELCOUNT",
    0x07e8: "UM_GETSELCOUNT",
    0x07e8: "WM_CPL_LAUNCH",
    0x07e9: "SM_GETSERVERSELA",
    0x07e9: "UM_GETUSERSELA",
    0x07e9: "WM_CPL_LAUNCHED",
    0x07ea: "SM_GETSERVERSELW",
    0x07ea: "UM_GETUSERSELW",
    0x07eb: "SM_GETCURFOCUSA",
    0x07eb: "UM_GETGROUPSELA",
    0x07ec: "SM_GETCURFOCUSW",
    0x07ec: "UM_GETGROUPSELW",
    0x07ed: "SM_GETOPTIONS",
    0x07ed: "UM_GETCURFOCUSA",
    0x07ee: "UM_GETCURFOCUSW",
    0x07ef: "UM_GETOPTIONS",
    0x07f0: "UM_GETOPTIONS2",
    0x2000: "OCMBASE",
    0x2019: "OCM_CTLCOLOR",
    0x202b: "OCM_DRAWITEM",
    0x202c: "OCM_MEASUREITEM",
    0x202d: "OCM_DELETEITEM",
    0x202e: "OCM_VKEYTOITEM",
    0x202f: "OCM_CHARTOITEM",
    0x2039: "OCM_COMPAREITEM",
    0x204e: "OCM_NOTIFY",
    0x2111: "OCM_COMMAND",
    0x2114: "OCM_HSCROLL",
    0x2115: "OCM_VSCROLL",
    0x2132: "OCM_CTLCOLORMSGBOX",
    0x2133: "OCM_CTLCOLOREDIT",
    0x2134: "OCM_CTLCOLORLISTBOX",
    0x2135: "OCM_CTLCOLORBTN",
    0x2136: "OCM_CTLCOLORDLG",
    0x2137: "OCM_CTLCOLORSCROLLBAR",
    0x2138: "OCM_CTLCOLORSTATIC",
    0x2210: "OCM_PARENTNOTIFY",
    0x8000: "WM_APP",
    0xcccd: "WM_RASDIALEVENT",
}



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stdint cimport uintptr_t

DEF NVJPEG_MAX_COMPONENT = 4

cdef extern from "library_types.h":
    cdef enum libraryPropertyType_t:
        MAJOR_VERSION
        MINOR_VERSION
        PATCH_LEVEL

ctypedef void* cudaStream_t
#cdef extern from "cuda_runtime_api.h":
#    ctypedef int cudaError_t
#    ctypedef void* cudaStream_t
#    cudaError_t cudaStreamCreate(cudaStream_t* pStream)
#    cudaError_t cudaStreamSynchronize(cudaStream_t stream)

cdef extern from "nvjpeg.h":
    int NVJPEG_MAX_COMPONENT

    int NVJPEG_VER_MAJOR    #ie: 11
    int NVJPEG_VER_MINOR    #ie: 3
    int NVJPEG_VER_PATCH    #ie: 1
    int NVJPEG_VER_BUILD    #ie: 68

    ctypedef void* NV_ENC_INPUT_PTR
    ctypedef void* NV_ENC_OUTPUT_PTR
    ctypedef void* NV_ENC_REGISTERED_PTR

    ctypedef enum nvjpegStatus_t:
        NVJPEG_STATUS_SUCCESS
        NVJPEG_STATUS_NOT_INITIALIZED
        NVJPEG_STATUS_INVALID_PARAMETER
        NVJPEG_STATUS_BAD_JPEG
        NVJPEG_STATUS_JPEG_NOT_SUPPORTED
        NVJPEG_STATUS_ALLOCATOR_FAILURE
        NVJPEG_STATUS_EXECUTION_FAILED
        NVJPEG_STATUS_ARCH_MISMATCH
        NVJPEG_STATUS_INTERNAL_ERROR
        NVJPEG_STATUS_IMPLEMENTATION_NOT_SUPPORTED

    ctypedef enum nvjpegChromaSubsampling_t:
        NVJPEG_CSS_444
        NVJPEG_CSS_422
        NVJPEG_CSS_420
        NVJPEG_CSS_440
        NVJPEG_CSS_411
        NVJPEG_CSS_410
        NVJPEG_CSS_GRAY
        NVJPEG_CSS_UNKNOWN

    ctypedef enum nvjpegOutputFormat_t:
        NVJPEG_OUTPUT_UNCHANGED
        # return planar luma and chroma, assuming YCbCr colorspace
        NVJPEG_OUTPUT_YUV
        # return luma component only, if YCbCr colorspace
        # or try to convert to grayscale,
        # writes to 1-st channel of nvjpegImage_t
        NVJPEG_OUTPUT_Y
        # convert to planar RGB
        NVJPEG_OUTPUT_RGB
        # convert to planar BGR
        NVJPEG_OUTPUT_BGR
        # convert to interleaved RGB and write to 1-st channel of nvjpegImage_t
        NVJPEG_OUTPUT_RGBI
        # convert to interleaved BGR and write to 1-st channel of nvjpegImage_t
        NVJPEG_OUTPUT_BGRI
        # maximum allowed value
        NVJPEG_OUTPUT_FORMAT_MAX

    ctypedef enum nvjpegInputFormat_t:
        NVJPEG_INPUT_RGB    # Input is RGB - will be converted to YCbCr before encoding
        NVJPEG_INPUT_BGR    # Input is RGB - will be converted to YCbCr before encoding
        NVJPEG_INPUT_RGBI   # Input is interleaved RGB - will be converted to YCbCr before encoding
        NVJPEG_INPUT_BGRI   # Input is interleaved RGB - will be converted to YCbCr before encoding

    ctypedef enum nvjpegBackend_t:
        NVJPEG_BACKEND_DEFAULT
        NVJPEG_BACKEND_HYBRID       # uses CPU for Huffman decode
        NVJPEG_BACKEND_GPU_HYBRID   # uses GPU assisted Huffman decode. nvjpegDecodeBatched will use GPU decoding for baseline JPEG bitstreams with
                                    # interleaved scan when batch size is bigger than 100
        NVJPEG_BACKEND_HARDWARE     # supports baseline JPEG bitstream with single scan. 410 and 411 sub-samplings are not supported

    ctypedef enum nvjpegJpegEncoding_t:
        NVJPEG_ENCODING_UNKNOWN
        NVJPEG_ENCODING_BASELINE_DCT
        NVJPEG_ENCODING_EXTENDED_SEQUENTIAL_DCT_HUFFMAN
        NVJPEG_ENCODING_PROGRESSIVE_DCT_HUFFMAN


    ctypedef struct nvjpegImage_t:
        unsigned char * channel[NVJPEG_MAX_COMPONENT]
        size_t    pitch[NVJPEG_MAX_COMPONENT]

    ctypedef struct nvjpegDevAllocator_t:
        pass

    ctypedef struct nvjpegPinnedAllocator_t:
        pass

    ctypedef struct nvjpegHandle:
        pass
    ctypedef nvjpegHandle* nvjpegHandle_t

    nvjpegStatus_t nvjpegGetProperty(libraryPropertyType_t type, int *value)
    nvjpegStatus_t nvjpegGetCudartProperty(libraryPropertyType_t type, int *value)
    nvjpegStatus_t nvjpegCreate(nvjpegBackend_t backend, nvjpegDevAllocator_t *dev_allocator, nvjpegHandle_t *handle)
    nvjpegStatus_t nvjpegCreateSimple(nvjpegHandle_t *handle)
    nvjpegStatus_t nvjpegCreateEx(nvjpegBackend_t backend,
        nvjpegDevAllocator_t *dev_allocator,
        nvjpegPinnedAllocator_t *pinned_allocator,
        unsigned int flags,
        nvjpegHandle_t *handle)

    nvjpegStatus_t nvjpegDestroy(nvjpegHandle_t handle)
    nvjpegStatus_t nvjpegSetDeviceMemoryPadding(size_t padding, nvjpegHandle_t handle)
    nvjpegStatus_t nvjpegGetDeviceMemoryPadding(size_t *padding, nvjpegHandle_t handle)
    nvjpegStatus_t nvjpegSetPinnedMemoryPadding(size_t padding, nvjpegHandle_t handle)
    nvjpegStatus_t nvjpegGetPinnedMemoryPadding(size_t *padding, nvjpegHandle_t handle)

    ctypedef struct nvjpegJpegState_t:
        pass
    nvjpegStatus_t nvjpegJpegStateCreate(nvjpegHandle_t handle, nvjpegJpegState_t *jpeg_handle)
    nvjpegStatus_t nvjpegJpegStateDestroy(nvjpegJpegState_t handle)

    nvjpegStatus_t nvjpegGetImageInfo(
        nvjpegHandle_t handle,
        const unsigned char *data,
        size_t length,
        int *nComponents,
        nvjpegChromaSubsampling_t *subsampling,
        int *widths,
        int *heights)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import math
from time import monotonic
from typing import Any
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.util.env import envint, envbool
from xpra.util.rectangle import rectangle, add_rectangle, remove_rectangle, merge_all
from xpra.log import Logger

sslog = Logger("regiondetect")
refreshlog = Logger("regionrefresh")

GLib = gi_import("GLib")

VIDEO_SUBREGION = envbool("XPRA_VIDEO_SUBREGION", True)
SUBWINDOW_REGION_BOOST = envint("XPRA_SUBWINDOW_REGION_BOOST", 20)

MAX_TIME = envint("XPRA_VIDEO_DETECT_MAX_TIME", 5)
MIN_EVENTS = envint("XPRA_VIDEO_DETECT_MIN_EVENTS", 20)
MIN_W = envint("XPRA_VIDEO_DETECT_MIN_WIDTH", 128)
MIN_H = envint("XPRA_VIDEO_DETECT_MIN_HEIGHT", 96)

RATIO_WEIGHT = envint("XPRA_VIDEO_DETECT_RATIO_WEIGHT", 80)
KEEP_SCORE = envint("XPRA_VIDEO_DETECT_KEEP_SCORE", 160)


def scoreinout(ww: int, wh: int, region, incount: int, outcount: int) -> int:
    total = incount + outcount
    assert total > 0
    # proportion of damage events that are within this region:
    inregion = incount / total
    # devaluate by taking into account the number of pixels in the area
    # so that a large video region only wins if it really
    # has a larger proportion of the pixels
    # (but also offset this value to even things out a bit:
    # if we have a series of vertical or horizontal bands that we merge,
    # we would otherwise end up excluding the ones on the edge
    # if they ever happen to have a slightly lower hit count)
    # summary: bigger is better, as long as we still have more pixels in than out
    width = min(ww, region.width)
    height = min(wh, region.height)
    # proportion of pixels in this region relative to the whole window:
    inwindow = (width * height) / (ww * wh)
    ratio = inregion / inwindow
    score = 100.0 * inregion
    # if the region has at least 35% of updates, boost it with window ratio
    # (capped at 6, and smoothed with sqrt):
    score += max(0.0, inregion - 0.35) * (math.sqrt(min(6.0, ratio)) - 1.0) * RATIO_WEIGHT
    sslog("scoreinout(%i, %i, %s, %i, %i) inregion=%i%%, inwindow=%i%%, ratio=%.1f, score=%i",
          ww, wh, region, incount, outcount, 100 * inregion, 100 * inwindow, ratio, score)
    return max(0, int(score))


class VideoSubregion:
    def __init__(self, refresh_cb: Callable, auto_refresh_delay: int, supported=False):
        self.refresh_cb = refresh_cb  # usage: refresh_cb(window, regions)
        self.auto_refresh_delay = auto_refresh_delay
        self.supported = supported
        self.enabled = True
        self.detection = True
        self.exclusion_zones: list[rectangle] = []
        self.init_vars()

    def init_vars(self) -> None:
        self.rectangle: rectangle | None = None
        self.inout = 0, 0  # number of damage pixels within / outside the region
        self.score = 0
        self.fps = 0
        self.damaged = 0  # proportion of the rectangle that got damaged (percentage)
        self.set_at = 0  # value of the "damage event count" when the region was set
        self.counter = 0  # value of the "damage event count" recorded at "time"
        self.time: float = 0  # see above
        self.refresh_timer = 0
        self.refresh_regions: list[rectangle] = []
        self.last_scores: dict[rectangle | None, int] = {}
        self.nonvideo_regions: list[rectangle] = []
        self.nonvideo_refresh_timer = 0
        # keep track of how much extra we batch non-video regions (milliseconds):
        self.non_max_wait = 150
        self.min_time = monotonic()

    def reset(self) -> None:
        self.cancel_refresh_timer()
        self.cancel_nonvideo_refresh_timer()
        self.init_vars()

    def cleanup(self) -> None:
        self.reset()

    def __repr__(self):
        return f"VideoSubregion({self.rectangle})"

    def set_enabled(self, enabled: bool) -> None:
        self.enabled = enabled
        if not enabled:
            self.novideoregion("disabled")

    def set_detection(self, detection: bool) -> None:
        self.detection = detection
        if not self.detection:
            self.reset()

    def set_region(self, x: int, y: int, w: int, h: int) -> None:
        sslog("set_region%s", (x, y, w, h))
        if self.detection:
            sslog("video region detection is on - the given region may or may not stick")
        if x == 0 and y == 0 and w == 0 and h == 0:
            self.novideoregion("empty")
        else:
            self.rectangle = rectangle(x, y, w, h)

    def set_exclusion_zones(self, zones) -> None:
        rects = []
        for (x, y, w, h) in zones:
            rects.append(rectangle(int(x), int(y), int(w), int(h)))
        self.exclusion_zones = rects
        # force expire:
        self.counter = 0

    def set_auto_refresh_delay(self, d: int) -> None:
        refreshlog("subregion auto-refresh delay: %s", d)
        if not isinstance(d, int):
            raise ValueError(f"delay is not an int: {d} ({type(d)})")
        self.auto_refresh_delay = d

    def cancel_refresh_timer(self) -> None:
        rt = self.refresh_timer
        refreshlog("%s.cancel_refresh_timer() timer=%s", self, rt)
        if rt:
            self.refresh_timer = 0
            GLib.source_remove(rt)

    def get_info(self) -> dict[str, Any]:
        r = self.rectangle
        info = {
            "supported": self.supported,
            "enabled": self.enabled,
            "detection": self.detection,
            "counter": self.counter,
            "auto-refresh-delay": self.auto_refresh_delay,
        }
        if r is None:
            return info
        info.update({
            "x": r.x,
            "y": r.y,
            "width": r.width,
            "height": r.height,
            "rectangle": (r.x, r.y, r.width, r.height),
            "set-at": self.set_at,
            "time": int(self.time),
            "min-time": int(self.min_time),
            "non-max-wait": self.non_max_wait,
            "timer": self.refresh_timer,
            "nonvideo-timer": self.nonvideo_refresh_timer,
            "in-out": self.inout,
            "score": self.score,
            "fps": self.fps,
            "damaged": self.damaged,
            "exclusion-zones": [(r.x, r.y, r.width, r.height) for r in self.exclusion_zones]
        })
        ls = self.last_scores
        if ls:
            # convert rectangles into tuples:
            info["scores"] = {r.get_geometry(): score for r, score in ls.items() if r is not None}
        rr = tuple(self.refresh_regions)
        if rr:
            for i, r in enumerate(rr):
                info[f"refresh_region[{i}]"] = (r.x, r.y, r.width, r.height)
        nvrr = tuple(self.nonvideo_regions)
        if nvrr:
            for i, r in enumerate(nvrr):
                info[f"nonvideo_refresh_region[{i}]"] = (r.x, r.y, r.width, r.height)
        return info

    def remove_refresh_region(self, region) -> None:
        remove_rectangle(self.refresh_regions, region)
        remove_rectangle(self.nonvideo_regions, region)
        refreshlog("remove_refresh_region(%s) updated refresh regions=%s, nonvideo regions=%s",
                   region, self.refresh_regions, self.nonvideo_regions)

    def add_video_refresh(self, region) -> None:
        # called by add_refresh_region if the video region got painted on
        # Note: this does not run in the UI thread!
        rect = self.rectangle
        if not rect:
            return
        # something in the video region is still refreshing,
        # so we re-schedule the subregion refresh:
        self.cancel_refresh_timer()
        # add the new region to what we already have:
        add_rectangle(self.refresh_regions, region)
        # do refresh any regions which are now outside the current video region:
        # (this can happen when the region moves or changes size)
        nonvideo = []
        for r in self.refresh_regions:
            if not rect.contains_rect(r):
                nonvideo += r.subtract_rect(rect)
        delay = max(150, self.auto_refresh_delay)
        self.nonvideo_regions += nonvideo
        if self.nonvideo_regions:
            if not self.nonvideo_refresh_timer:
                # refresh via timeout_add so this will run in the UI thread:
                self.nonvideo_refresh_timer = GLib.timeout_add(delay, self.nonvideo_refresh)
            # only keep the regions still in the video region:
            inrect = (rect.intersection_rect(r) for r in self.refresh_regions)
            self.refresh_regions = [r for r in inrect if r is not None]
        refreshlog("add_video_refresh(%s) rectangle=%s, delay=%ims, nonvideo=%s, refresh_regions=%s",
                   region, rect, delay, self.nonvideo_regions, self.refresh_regions)
        # re-schedule the video region refresh (if we still have regions to fresh):
        if self.refresh_regions:
            self.refresh_timer = GLib.timeout_add(delay, self.refresh)

    def cancel_nonvideo_refresh_timer(self) -> None:
        nvrt = self.nonvideo_refresh_timer
        refreshlog("cancel_nonvideo_refresh_timer() timer=%s", nvrt)
        if nvrt:
            self.nonvideo_refresh_timer = 0
            GLib.source_remove(nvrt)
            self.nonvideo_regions = []

    def nonvideo_refresh(self) -> None:
        self.nonvideo_refresh_timer = 0
        nonvideo = tuple(self.nonvideo_regions)
        refreshlog("nonvideo_refresh() nonvideo regions=%s", nonvideo)
        if not nonvideo:
            return
        if self.refresh_cb(nonvideo):
            self.nonvideo_regions = []
        # if the refresh didn't fire (refresh_cb() returned False),
        # then we should end up re-scheduling the nonvideo refresh
        # from add_video_refresh()

    def refresh(self) -> None:
        regions = self.refresh_regions
        rect = self.rectangle
        refreshlog("refresh() refresh_timer=%s, refresh_regions=%s, rectangle=%s",
                   self.refresh_timer, regions, rect)
        # runs via timeout_add, safe to call UI!
        self.refresh_timer = 0
        if rect and len(regions) >= 2:
            # figure out if it makes sense to refresh the whole area,
            # or if we just send the list of smaller rectangles:
            pixels = sum(r.width * r.height for r in regions)
            if pixels >= rect.width * rect.height // 2:
                regions = [rect]
        refreshlog("refresh() calling %s with regions=%s", self.refresh_cb, regions)
        if self.refresh_cb(regions):
            self.refresh_regions = []
        else:
            # retry later
            self.refresh_timer = GLib.timeout_add(1000, self.refresh)

    def novideoregion(self, msg, *args) -> None:
        sslog("novideoregion: " + msg, *args)
        self.rectangle = None
        self.time = 0
        self.set_at = 0
        self.counter = 0
        self.inout = 0, 0
        self.score = 0
        self.fps = 0
        self.damaged = 0

    def excluded_rectangles(self, rect, ww: int, wh: int) -> list:
        rects = [rect]
        if self.exclusion_zones:
            for e in self.exclusion_zones:
                new_rects = []
                for r in rects:
                    ex, ey, ew, eh = e.get_geometry()
                    if ex < 0 or ey < 0:
                        # negative values are relative to the width / height of the window:
                        if ex < 0:
                            ex = max(0, ww - ew)
                        if ey < 0:
                            ey = max(0, wh - eh)
                    new_rects += r.subtract(ex, ey, ew, eh)
                rects = new_rects
        return rects

    def identify_video_subregion(self, ww: int, wh: int, damage_events_count, last_damage_events,
                                 starting_at=0.0, children=()):
        if not self.enabled or not self.supported:
            self.novideoregion("disabled")
            return
        from_time: float = 0
        if not self.detection:
            if not self.rectangle:
                return
            # just update the fps:
            from_time = max(starting_at, monotonic() - MAX_TIME, self.min_time)
            self.time = monotonic()
            lde = tuple(x for x in tuple(last_damage_events) if x[0] >= from_time)
            incount = 0
            for _, x, y, w, h in lde:
                r = rectangle(x, y, w, h)
                inregion = r.intersection_rect(self.rectangle)
                if inregion:
                    incount += inregion.width * inregion.height
            elapsed = monotonic() - from_time
            if elapsed <= 0:
                self.fps = 0
            else:
                self.fps = int(incount / (self.rectangle.width * self.rectangle.height) / elapsed)
            return
        sslog("%s.identify_video_subregion(..)", self)
        sslog("identify_video_subregion%s",
              (ww, wh, damage_events_count, last_damage_events, starting_at, children))

        children_rects: Sequence[rectangle] = ()
        if children:
            children_rects = tuple(rectangle(x, y, w, h)
                                   for _xid, x, y, w, h, _border, _depth in children
                                   if w >= MIN_W and h >= MIN_H)

        if damage_events_count < self.set_at:
            # stats got reset
            self.set_at = 0
        # validate against window dimensions:
        rect = self.rectangle
        if rect and (rect.width > ww or rect.height > wh):
            # region is now bigger than the window!
            self.novideoregion("window is now smaller than current region")
            return
        # arbitrary minimum size for regions we will look at:
        # (we don't want video regions smaller than this - too much effort for little gain)
        if ww < MIN_W or wh < MIN_H:
            self.novideoregion("window is too small: %sx%s", MIN_W, MIN_H)
            return

        def update_markers():
            self.counter = damage_events_count
            self.time = monotonic()

        if self.counter + 10 > damage_events_count:
            # less than 10 events since last time we called update_markers:
            elapsed = monotonic() - self.time
            # how many damage events occurred since we chose this region:
            event_count = max(0, damage_events_count - self.set_at)
            # make the timeout longer when the region has worked longer:
            slow_region_timeout = 2 + math.log(2 + event_count, 1.5)
            if rect and elapsed >= slow_region_timeout:
                update_markers()
                self.novideoregion("too much time has passed (%is for %i total events)", elapsed, event_count)
                return
            sslog("identify video: waiting for more damage events (%i) counters: %i / %i",
                  event_count, self.counter, damage_events_count)
            return

        from_time = max(starting_at, monotonic() - MAX_TIME, self.min_time)
        # create a list (copy) to work on:
        lde = tuple(x for x in tuple(last_damage_events) if x[0] >= from_time)
        dc = len(lde)
        if dc <= MIN_EVENTS:
            self.novideoregion("not enough damage events yet (%s)", dc)
            return
        # structures for counting areas and sizes:
        wc: dict[int, dict[int, set[rectangle]]] = {}
        hc: dict[int, dict[int, set[rectangle]]] = {}
        dec: dict[rectangle, int] = {}
        # count how many times we see each area, each width/height and where,
        # after removing any exclusion zones:
        for _, x, y, w, h in lde:
            rects = self.excluded_rectangles(rectangle(x, y, w, h), ww, wh)
            for r in rects:
                dec[r] = dec.get(r, 0) + 1
                if w >= MIN_W:
                    wc.setdefault(w, {}).setdefault(x, set()).add(r)
                if h >= MIN_H:
                    hc.setdefault(h, {}).setdefault(y, set()).add(r)
        # we can shortcut the damaged ratio if the whole window got damaged at least once:
        all_damaged = dec.get(rectangle(0, 0, ww, wh), 0) > 0

        def inoutcount(region, ignore_size=0):
            # count how many pixels are in or out if this region
            incount, outcount = 0, 0
            for r, count in dec.items():
                inregion = r.intersection_rect(region)
                if inregion:
                    incount += inregion.width * inregion.height * int(count)
                outregions = r.subtract_rect(region)
                for x in outregions:
                    if ignore_size > 0 and x.width * x.height < ignore_size:
                        # skip small region outside rectangle
                        continue
                    outcount += x.width * x.height * int(count)
            return incount, outcount

        def damaged_ratio(rect: rectangle):
            if all_damaged:
                return 1
            rects: list[rectangle] = [rect, ]
            for _, x, y, w, h in lde:
                r = rectangle(x, y, w, h)
                new_rects = []
                for cr in rects:
                    new_rects += cr.subtract_rect(r)
                if not new_rects:
                    # nothing left: damage covered the whole rect
                    return 1.0
                rects = new_rects
            not_damaged_pixels = sum((r.width * r.height) for r in rects)
            rect_pixels = rect.width * rect.height
            # sslog("damaged_ratio: not damaged pixels(%s)=%i, rect pixels(%s)=%i",
            #     rects, not_damaged_pixels, rect, rect_pixels)
            return max(0, min(1, 1.0 - not_damaged_pixels / rect_pixels))

        scores = {None: 0}

        def score_region(info: str, region: rectangle, ignore_size=0, d_ratio=0.0) -> int:
            score = scores.get(region)
            if score is not None:
                return score

            def rec(score: int):
                scores[region] = score
                return score

            # check if the region given is a good candidate, and if so we use it
            # clamp it:
            if region.width < MIN_W or region.height < MIN_H:
                # too small, ignore it:
                sslog(f"region too small: {region.width}x{region.height}")
                return rec(0)
            # and make sure this does not end up much bigger than needed:
            if ww * wh < (region.width * region.height):
                sslog(f"region too small: {region.width}x{region.height}")
                return rec(0)
            incount, outcount = inoutcount(region, ignore_size)
            total = incount + outcount
            children_boost = 0
            if total == 0:
                ipct = opct = score = 0
            else:
                ipct = 100 * incount // total
                opct = 100 * outcount // total
            if score is None:
                score = scoreinout(ww, wh, region, incount, outcount)
                # discount score if the region contains areas that were not damaged:
                # (apply sqrt to limit the discount: 50% damaged -> multiply by 0.7)
                if d_ratio == 0:
                    d_ratio = damaged_ratio(region)
                score = int(score * math.sqrt(d_ratio))
                children_boost = int(region in children_rects) * SUBWINDOW_REGION_BOOST
            sslog("testing %12s video region %34s: "
                  "%3i%% in, %3i%% out, %3i%% of window, damaged ratio=%.2f, children_boost=%i, score=%2i",
                  info, region, ipct, opct, 100 * region.width * region.height / ww / wh, d_ratio, children_boost,
                  score)
            return rec(score)

        def updateregion(rect: rectangle) -> None:
            self.rectangle = rect
            self.time = monotonic()
            self.inout = inoutcount(rect)
            self.score = scoreinout(ww, wh, rect, *self.inout)
            elapsed = monotonic() - from_time
            if elapsed <= 0:
                self.fps = 0
            else:
                self.fps = int(self.inout[0] / (rect.width * rect.height) / elapsed)
            self.damaged = int(100 * damaged_ratio(self.rectangle))
            self.last_scores = scores
            sslog("score(%s)=%s, damaged=%i%%", self.inout, self.score, self.damaged)

        def setnewregion(rect: rectangle, msg: str, *args) -> None:
            rects = self.excluded_rectangles(rect, ww, wh)
            if not rects:
                self.novideoregion("no match after removing excluded regions")
                return
            if len(rects) == 1:
                rect = rects[0]
            else:
                # use the biggest one of what remains:
                def get_rect_size(rect):
                    return -rect.width * rect.height

                biggest_rects = sorted(rects, key=get_rect_size)
                rect = biggest_rects[0]
                if rect.width < MIN_W or rect.height < MIN_H:
                    self.novideoregion("match is too small after removing excluded regions")
                    return
            if not self.rectangle or self.rectangle != rect:
                sslog("setting new region %s: " + msg, rect, *args)
                sslog(" is child window: %s", rect in children_rects)
                self.set_at = damage_events_count
                self.counter = damage_events_count
            if not self.enabled:
                # could have been disabled since we started this method!
                self.novideoregion("disabled")
                return
            if not self.detection:
                return
            updateregion(rect)

        update_markers()

        if len(dec) == 1:
            rect, count = tuple(dec.items())[0]
            setnewregion(rect, "only region damaged")
            return

        # see if we can keep the region we already have (if any):
        cur_score = 0
        if rect:
            cur_score = score_region("current", rect)
            if cur_score >= KEEP_SCORE:
                sslog("keeping existing video region %s with score %s", rect, cur_score)
                return

        # split the regions we really care about (enough pixels, big enough):
        damage_count = {}
        min_count = max(2, len(lde) // 40)
        for r, count in dec.items():
            # ignore small regions:
            if count > min_count and r.width >= MIN_W and r.height >= MIN_H:
                damage_count[r] = count
        c = sum(int(x) for x in damage_count.values())
        if c > 0:
            most_damaged = int(sorted(damage_count.values())[-1])
            most_pct = round(100 * most_damaged / c)
            sslog("identify video: most=%s%% damage count=%s", most_pct, damage_count)
            # is there a region that stands out?
            # try to use the region which is responsible for most of the large damage requests:
            most_damaged_regions = tuple(r for r, v in damage_count.items() if v == most_damaged)
            if len(most_damaged_regions) == 1:
                r = most_damaged_regions[0]
                score = score_region("most-damaged", r, d_ratio=1.0)
                sslog(f"identify video: score most damaged area {r}={score}%")
                if score > 120:
                    setnewregion(r, f"{most_pct}% of large damage requests, {score=}")
                    return
                if score >= 100:
                    scores[r] = score

        # try children windows:
        for region in children_rects:
            scores[region] = score_region("child-window", region, 48 * 48)

        # try harder: try combining regions with the same width or height:
        # (some video players update the video region in bands)
        for w, d in wc.items():
            for x, regions in d.items():
                if len(regions) >= 2:
                    # merge regions of width w at x
                    min_count = max(2, len(regions) // 25)
                    keep = tuple(r for r in regions if int(dec.get(r, 0)) >= min_count)
                    sslog("vertical regions of width %i at %i with at least %i hits: %s", w, x, min_count, keep)
                    if keep:
                        merged = merge_all(keep)
                        scores[merged] = score_region("vertical", merged, 48 * 48)
        for h, d in hc.items():
            for y, regions in d.items():
                if len(regions) >= 2:
                    # merge regions of height h at y
                    min_count = max(2, len(regions) // 25)
                    keep = tuple(r for r in regions if int(dec.get(r, 0)) >= min_count)
                    sslog("horizontal regions of height %i at %i with at least %i hits: %s", h, y, min_count, keep)
                    if keep:
                        merged = merge_all(keep)
                        scores[merged] = score_region("horizontal", merged, 48 * 48)

        sslog("merged regions scores: %s", scores)
        highscore = max(scores.values())
        # a score of 100 is neutral
        if highscore >= 120:
            region = next(iter(r for r, s in scores.items() if s == highscore))
            setnewregion(region, "very high score: %s", highscore)
            return

        # retry existing region, tolerate lower score:
        if cur_score >= 90 and (highscore < 100 or cur_score >= highscore):
            sslog("keeping existing video region %s with score %s", rect, cur_score)
            setnewregion(self.rectangle, f"existing region with score: {cur_score}")
            return

        if highscore >= 100:
            region = next(iter(r for r, s in scores.items() if s == highscore))
            setnewregion(region, "high score: %s", highscore)
            return

        # could do:
        # * re-add some scrolling detection: the region may have moved
        # * re-try with a higher "from_time" and a higher score threshold

        # try harder still: try combining all the regions we haven't discarded
        # (Flash player with Firefox and Youtube does stupid unnecessary repaints)
        if len(damage_count) >= 2:
            merged = merge_all(tuple(damage_count.keys()))
            score = score_region("merged", merged)
            if score >= 110:
                setnewregion(merged, "merged all regions, score=%s", score)
                return

        self.novideoregion("failed to identify a video region")
        self.last_scores = scores



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path

from xpra.os_util import gi_import
from xpra.log import Logger

GdkPixbuf = gi_import("GdkPixbuf")

log = Logger("gtk", "util")


def get_icon_from_file(filename) -> GdkPixbuf.Pixbuf | None:
    if not filename:
        log("get_icon_from_file(%s)=None", filename)
        return None
    try:
        if not os.path.exists(filename):
            log.warn("Warning: cannot load icon, '%s' does not exist", filename)
            return None
        with open(filename, mode="rb") as f:
            data = f.read()
        loader = GdkPixbuf.PixbufLoader()
        loader.write(data)
        loader.close()
    except Exception as e:
        log("get_icon_from_file(%s)", filename, exc_info=True)
        log.error("Error: failed to load '%s'", filename)
        log.estr(e)
        return None
    pixbuf = loader.get_pixbuf()
    return pixbuf


def get_icon_pixbuf(icon_name: str) -> GdkPixbuf.Pixbuf | None:
    if not icon_name:
        log("get_icon_pixbuf(%s)=None", icon_name)
        return None
    with log.trap_error("Error loading icon pixbuf %s", icon_name):
        from xpra.platform.paths import get_icon_filename
        icon_filename = get_icon_filename(icon_name)
        log("get_pixbuf(%s) icon_filename=%s", icon_name, icon_filename)
        if icon_filename:
            return GdkPixbuf.Pixbuf.new_from_file(filename=icon_filename)
    return None


def get_pixbuf_from_data(rgb_data, has_alpha: bool, w: int, h: int, rowstride: int) -> GdkPixbuf.Pixbuf:
    glib = gi_import("GLib")
    data = glib.Bytes(rgb_data)
    return GdkPixbuf.Pixbuf.new_from_bytes(data, GdkPixbuf.Colorspace.RGB,
                                           has_alpha, 8, w, h, rowstride)


def pixbuf_save_to_memory(pixbuf, fmt="png") -> bytes:
    buf = []

    def save_to_memory(data, *_args, **_kwargs):
        buf.append(data)
        return True

    pixbuf.save_to_callbackv(save_to_memory, None, fmt, [], [])
    return b"".join(buf)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable

from xpra.common import noop
from xpra.os_util import gi_import
from xpra.net.common import may_log_packet
from xpra.net.common import ServerPacketHandlerType
from xpra.log import Logger

log = Logger("server")

GLib = gi_import("GLib")


class GLibPacketHandler:

    def __init__(self):
        self._authenticated_packet_handlers: dict[str, Callable] = {}
        self._authenticated_ui_packet_handlers: dict[str, Callable] = {}
        self._default_packet_handlers: dict[str, Callable] = {}
        self.packet_alias: dict[str, str] = {}

    def get_info(self) -> dict[str, Any]:
        return {
            "packet-handlers" : {
                "authenticated": sorted(self._authenticated_packet_handlers.keys()),
                "ui": sorted(self._authenticated_ui_packet_handlers.keys()),
            },
        }

    def remove_packet_handlers(self, *keys) -> None:
        for k in keys:
            for d in (
                    self._authenticated_packet_handlers,
                    self._authenticated_ui_packet_handlers,
                    self._default_packet_handlers,
            ):
                d.pop(k, None)

    def add_packet_handlers(self, defs: dict[str, ServerPacketHandlerType], main_thread=False) -> None:
        for packet_type, handler in defs.items():
            self.add_packet_handler(packet_type, handler, main_thread)

    def add_packet_handler(self, packet_type: str, handler: ServerPacketHandlerType, main_thread=False) -> None:
        log("add_packet_handler%s", (packet_type, handler, main_thread))
        # replace any previously defined handlers:
        self.remove_packet_handlers(packet_type)
        if main_thread:
            handlers = self._authenticated_ui_packet_handlers
        else:
            handlers = self._authenticated_packet_handlers
        handlers[packet_type] = handler

    def add_packets(self, *packet_types: str, main_thread=False) -> None:
        for packet_type in packet_types:
            handler = getattr(self, "_process_" + packet_type.replace("-", "_"))
            self.add_packet_handler(packet_type, handler, main_thread)

    def add_legacy_alias(self, legacy_name, new_name) -> None:
        self.packet_alias[legacy_name] = new_name

    def dispatch_packet(self, proto, packet, authenticated=False) -> None:
        packet_type = str(packet[0])
        packet_type = self.packet_alias.get(packet_type, packet_type)
        handler: Callable = noop

        def call_handler() -> None:
            may_log_packet(False, packet_type, packet)
            self.call_packet_handler(handler, proto, packet)

        try:

            if authenticated:
                handler = self._authenticated_ui_packet_handlers.get(packet_type)
                if handler:
                    log("process ui packet %s", packet_type)
                    GLib.idle_add(call_handler)
                    return
                handler = self._authenticated_packet_handlers.get(packet_type)
                if handler:
                    log("process non-ui packet %s", packet_type)
                    call_handler()
                    return
            handler = self._default_packet_handlers.get(packet_type)
            if handler:
                log("process default packet %s", packet_type)
                call_handler()
                return

            GLib.idle_add(self.handle_invalid_packet, proto, packet)
        except (RuntimeError, AssertionError):
            log.error(f"Error processing a {packet_type!r} packet")
            log.error(f" received from {proto}:")
            log.error(f" using {handler}", exc_info=True)

    @staticmethod
    def call_packet_handler(handler: Callable, proto, packet) -> None:
        handler(proto, packet)

    @staticmethod
    def handle_invalid_packet(proto, packet) -> None:
        if proto.is_closed():
            return
        packet_type = str(packet[0])
        log("invalid packet: %s", packet)
        log.error(f"Error: unknown or invalid packet type {packet_type!r}")
        log.error(f" received from {proto}")
        proto.close()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import threading
from collections.abc import Callable
from threading import Thread
from typing import Sequence, Any

from xpra.util.io import get_util_logger
from xpra.util.system import nn
from xpra.util.thread import main_thread


def dump_all_frames(logger=None) -> None:
    try:
        frames = sys._current_frames()  # pylint: disable=protected-access
    except AttributeError:
        return
    else:
        dump_frames(frames.items(), logger)


def dump_gc_frames(logger=None) -> None:
    import gc
    import inspect
    gc.collect()
    frames = tuple((None, x) for x in gc.get_objects() if inspect.isframe(x))
    dump_frames(frames, logger)


def dump_frames(frames, logger=None) -> None:
    if not logger:
        from xpra.util.io import get_util_logger
        logger = get_util_logger()
    logger("found %s frames:", len(frames))
    for i, (fid, frame) in enumerate(frames):
        fidstr = ""
        if fid is not None:
            try:
                fidstr = hex(fid)
            except TypeError:
                fidstr = str(fid)
        logger("%i: %s", i, fidstr, frame=frame)


def detect_leaks() -> Callable[[], bool]:
    import tracemalloc
    tracemalloc.start()
    last_snapshot = [tracemalloc.take_snapshot()]

    def print_leaks() -> bool:
        s1 = last_snapshot[0]
        s2 = tracemalloc.take_snapshot()
        last_snapshot[0] = s2
        top_stats = s2.compare_to(s1, 'lineno')
        print("[ Top 20 differences ]")
        for stat in top_stats[:20]:
            print(stat)
        for i, stat in enumerate(top_stats[:20]):
            print()
            print("top %i:" % i)
            print("{} memory blocks: {:.1f} KiB".format(stat.count, stat.size / 1024))
            for line in stat.traceback.format():
                print(line)
        return True

    return print_leaks


def start_mem_watcher(ms) -> None:
    from xpra.util.thread import start_thread
    start_thread(mem_watcher, name="mem-watcher", daemon=True, args=(ms,))


def mem_watcher(ms, pid: int = os.getpid()) -> None:
    import time
    import psutil
    process = psutil.Process(pid)
    from xpra.util.io import get_util_logger
    logger = get_util_logger()
    while True:
        mem = process.memory_full_info()
        logger.info("memory usage for %s: %s", pid, mem)
        time.sleep(ms / 1000.0)


def log_mem_info(prefix="memory usage: ", pid=os.getpid()) -> None:
    import psutil
    process = psutil.Process(pid)
    mem = process.memory_full_info()
    print("%i %s%s" % (pid, prefix, mem))


def enforce_features(features, feature_map: dict[str, str]) -> None:
    """
    Prevent the modules from being imported later
    """
    from xpra.util.env import envbool
    debug_features = envbool("XPRA_FEATURES_DEBUG", False)
    for feature, modules in feature_map.items():
        enabled: bool = getattr(features, feature)
        for module in modules.split(","):
            value = sys.modules.get(module)
            if debug_features:
                from importlib.util import find_spec
                try:
                    exists = find_spec(module)
                except ModuleNotFoundError:
                    exists = False
                from xpra.util.io import get_util_logger
                log = get_util_logger()
                log.info(f"feature {feature!r:20}: {module!r:40} {enabled=}, found={exists}, value={value}")
            if not enabled:
                if value is not None:
                    from xpra.util.io import get_util_logger
                    log = get_util_logger()
                    log.warn(f"Warning: cannot disable {feature!r} feature")
                    log.warn(f" the module {module!r} is already loaded")
                else:
                    # noinspection PyTypeChecker
                    sys.modules[module] = None


def get_frame_info(ignore_threads: Sequence[Thread] = ()) -> dict[str | int, Any]:
    info: dict[str | int, Any] = {
        "count": threading.active_count() - len(ignore_threads),
        "native-id": threading.get_native_id(),
    }
    try:
        import traceback
        thread_ident: dict[int | None, str | None] = {}
        for t in threading.enumerate():
            if t not in ignore_threads:
                thread_ident[t.ident] = t.name
            else:
                thread_ident[t.ident] = None
        thread_ident |= {
            threading.current_thread().ident: "info",
            main_thread.ident: "main",
        }
        frames = sys._current_frames()  # pylint: disable=protected-access
        stack = None
        for i, frame_pair in enumerate(frames.items()):
            stack = traceback.extract_stack(frame_pair[1])
            tident = thread_ident.get(frame_pair[0], "unknown")
            if tident is None:
                continue
            # sanitize stack to prevent None values (which cause encoding errors with the bencoder)
            sanestack = []
            for entry in stack:
                sanestack.append(tuple(nn(x) for x in entry))
            del stack
            info[i] = {
                "": tident,
                "stack": sanestack,
            }
        del frames
    except Exception as e:
        get_util_logger().error("failed to get frame info: %s", e)
    return info



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

# pylint: disable=import-outside-toplevel

class TestDisplayUtil(unittest.TestCase):

    def test_repr(self):
        from xpra.x11.common import X11Event, REPR_FUNCTIONS
        class Custom():  # pylint: disable=too-few-public-methods
            def repr(self):
                return "Custom"
        def custom_repr(*_args):
            return "XXXXX"
        REPR_FUNCTIONS[Custom] = custom_repr
        name = "00name00"
        e = X11Event(name)
        e.display = "00display00"
        e.type = "00type00"
        e.serial = 255
        e.custom = Custom()
        def f(s, find=True):
            if repr(e).find(s)>=0 == find:
                #print("repr=%s" % repr(e))
                raise ValueError(f"{s!r} in {e!r}: {not find}")
        f(name)
        f(f"{e.serial:x}")
        f(e.display, False)
        f(e.type, False)
        f("XXXXX", True)


def main():
    from xpra.os_util import POSIX, OSX
    #can only work with an X11 server
    if POSIX and not OSX:
        unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import signal
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.exit_codes import ExitValue
from xpra.platform.gui import init as gui_init, force_focus
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import scaled_image, label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.log import Logger, consume_verbose_argv

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("util")


# noinspection PyTestUnpassedFixture
class UpdateStatusWindow:

    def __init__(self):
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.close)
        self.window.set_default_size(400, 200)
        self.window.set_title("Xpra Version Check")

        icon = get_icon_pixbuf("update.png")
        if icon:
            self.window.set_icon(icon)
        self.window.set_position(Gtk.WindowPosition.CENTER)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(0)

        # Label:
        self.progress = 0
        self.label = label("Version Check", font="sans 14")
        al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0)
        al.add(self.label)
        vbox.add(al)

        # Buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox)

        def btn(label: str, tooltip: str, callback: Callable, icon_name: str = "") -> Gtk.Button:
            btn = Gtk.Button(label=label)
            btn.set_tooltip_text(tooltip)
            btn.connect("clicked", callback)
            icon = get_icon_pixbuf(icon_name)
            if icon:
                btn.set_image(scaled_image(icon, 24))
            hbox.pack_start(btn)
            return btn

        btn("Download", "Show download page", self.download, "download.png")
        btn("Close", "", self.close, "quit.png")

        def accel_close(*_args) -> None:
            self.close()

        add_close_accel(self.window, accel_close)
        vbox.show_all()
        self.window.vbox = vbox
        self.window.add(vbox)
        self.newer_version: None | bool | Sequence[int] = None

    def check(self) -> None:
        if self.progress:
            return
        self.newer_version = None
        GLib.timeout_add(1000, self.update_label)
        from xpra.util.thread import start_thread
        start_thread(self.do_check, "version check", daemon=True)

    def do_check(self) -> None:
        from xpra.util.version import version_update_check
        try:
            self.newer_version = version_update_check()
        finally:
            self.progress = 0

    def update_label(self) -> bool:
        if self.newer_version is False:
            from xpra import __version__ as version_str
            self.label.set_label("Version %s is up to date" % version_str)
            return False
        if self.newer_version:
            version_str = ".".join(str(x) for x in self.newer_version)
            self.label.set_label("A newer version is available: %s" % version_str)
            return False
        self.label.set_label("Checking for new versions %s" % (["-", "\\", "|", "/"][self.progress % 4]))
        self.progress += 1
        return True

    def show(self) -> None:
        log("show()")

        def show() -> None:
            force_focus()
            self.window.show()
            self.window.present()

        GLib.idle_add(show)

    def hide(self) -> None:
        log("hide()")
        self.window.hide()

    def close(self, *args) -> bool:
        log("close%s", args)
        self.hide()
        return True

    def destroy(self, *args) -> None:
        log("destroy%s", args)
        if self.window:
            self.window.destroy()
            self.window = None

    def run(self) -> ExitValue:
        log("run()")
        Gtk.main()
        log("run() Gtk.main done")
        return 0

    def quit(self, *args) -> None:
        log("quit%s", args)
        self.hide()
        Gtk.main_quit()

    def download(self, *_args) -> None:
        self.hide()
        import webbrowser
        webbrowser.open_new_tab("https://github.com/Xpra-org/xpra/wiki/Download")


_instance = None


def get_update_status_window() -> UpdateStatusWindow:
    global _instance
    if _instance is None:
        _instance = UpdateStatusWindow()
    return _instance


def main(argv) -> int:
    from xpra.platform import program_context
    from xpra.platform.gui import ready as gui_ready
    with program_context("Xpra-Version-Check", "Xpra Version Check"):
        consume_verbose_argv(argv, "all")

        from xpra.gtk.signals import register_os_signals
        app = UpdateStatusWindow()
        app.close = app.quit
        register_os_signals(app.quit, "Version Check")
        try:
            gui_ready()
            app.show()
            app.check()
            return app.run()
        except KeyboardInterrupt:
            return 128 + int(signal.SIGINT)


if __name__ == "__main__":
    gui_init()
    v = main(sys.argv)
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.client.gtk3.opengl.client_window import GLClientWindowBase
from xpra.log import Logger

log = Logger("opengl", "paint")

Gtk = gi_import("Gtk")
GObject = gi_import("GObject")


class GLClientWindow(GLClientWindowBase):

    def get_backing_class(self):
        from xpra.client.gtk3.opengl.glarea_backing import GLAreaBacking
        return GLAreaBacking

    def repaint(self, x: int, y: int, w: int, h: int) -> None:
        widget = self.drawing_area
        log(f"repaint%s {widget=}", (x, y, w, h))
        if widget:
            widget.queue_render()


GObject.type_register(GLClientWindow)


def check_support(force_enable=False) -> dict[str, Any]:
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_default_size(400, 400)
    window.resize(400, 400)
    window.set_decorated(False)
    window.realize()
    from xpra.client.gtk3.opengl.glarea_backing import GLArea
    glarea = GLArea(True)
    from xpra.gtk.window import set_visual
    set_visual(glarea, True)
    window.add(glarea)
    glarea.realize()
    gl_context = glarea.get_context()
    gl_context.make_current()
    try:
        from xpra.opengl.check import check_PyOpenGL_support
        return check_PyOpenGL_support(force_enable)
    finally:
        window.destroy()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# @PydevCodeAnalysisIgnore

from ctypes import (
    WinDLL,  # @UnresolvedImport
    POINTER, Structure, Union, c_void_p, c_ubyte, addressof,
)
from ctypes.wintypes import DWORD, ULONG, HANDLE, BOOL, INT, BYTE, WORD, LPCSTR

from xpra.platform.win32.constants import WAIT_ABANDONED, WAIT_OBJECT_0, WAIT_TIMEOUT, WAIT_FAILED
from xpra.platform.win32.common import LPSECURITY_ATTRIBUTES

PDWORD = POINTER(DWORD)
PHANDLE = POINTER(HANDLE)
LPDWORD = POINTER(DWORD)
PVOID = c_void_p
LPCVOID = c_void_p
LPVOID = c_void_p

WAIT_STR = {
    WAIT_ABANDONED: "ABANDONED",
    WAIT_OBJECT_0: "OBJECT_0",
    WAIT_TIMEOUT: "TIMEOUT",
    WAIT_FAILED: "FAILED",
}

INFINITE = 65535
INVALID_HANDLE_VALUE = HANDLE(-1).value

INVALID_HANDLE = HANDLE(INVALID_HANDLE_VALUE)


class _inner_struct(Structure):
    _fields_ = [
        ('Offset', DWORD),
        ('OffsetHigh', DWORD),
    ]


class _inner_union(Union):
    _fields_ = [
        ('anon_struct', _inner_struct),
        ('Pointer', c_void_p),
    ]


class OVERLAPPED(Structure):
    _fields_ = [
        ('Internal', POINTER(ULONG)),
        ('InternalHigh', POINTER(ULONG)),
        ('union', _inner_union),
        ('hEvent', HANDLE),
    ]


LPOVERLAPPED = POINTER(OVERLAPPED)

kernel32 = WinDLL("kernel32", use_last_error=True)
WaitForSingleObject = kernel32.WaitForSingleObject
WaitForSingleObject.argtypes = [HANDLE, DWORD]
WaitForSingleObject.restype = DWORD
CreateEventA = kernel32.CreateEventA
CreateEventA.restype = HANDLE
SetEvent = kernel32.SetEvent
SetEvent.argtypes = [HANDLE]
SetEvent.restype = BOOL
ReadFile = kernel32.ReadFile
ReadFile.argtypes = [HANDLE, LPVOID, DWORD, LPDWORD, LPOVERLAPPED]
ReadFile.restype = BOOL
WriteFile = kernel32.WriteFile
WriteFile.argtypes = [HANDLE, LPCVOID, DWORD, LPDWORD, LPOVERLAPPED]
WriteFile.restype = BOOL
CreateFileA = kernel32.CreateFileA
CreateFileA.argtypes = [LPCSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES, DWORD, DWORD, HANDLE]
CreateFileA.restype = HANDLE
WaitNamedPipeA = kernel32.WaitNamedPipeA
SetNamedPipeHandleState = kernel32.SetNamedPipeHandleState
SetNamedPipeHandleState.argtypes = [HANDLE, LPDWORD, LPDWORD, LPDWORD]
SetNamedPipeHandleState.restype = INT
GetOverlappedResult = kernel32.GetOverlappedResult
GetOverlappedResult.argtypes = [HANDLE, LPOVERLAPPED, LPDWORD, BOOL]
GetOverlappedResult.restype = BOOL
CreateNamedPipeA = kernel32.CreateNamedPipeA
CreateNamedPipeA.argtypes = [LPCSTR, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, LPSECURITY_ATTRIBUTES]
CreateNamedPipeA.restype = HANDLE
ConnectNamedPipe = kernel32.ConnectNamedPipe
ConnectNamedPipe.argtypes = [HANDLE, LPOVERLAPPED]
ConnectNamedPipe.restype = BOOL
DisconnectNamedPipe = kernel32.DisconnectNamedPipe
DisconnectNamedPipe.argtypes = [HANDLE]
DisconnectNamedPipe.restype = BOOL
FlushFileBuffers = kernel32.FlushFileBuffers
FlushFileBuffers.argtypes = [HANDLE]
FlushFileBuffers.restype = BOOL
GetLastError = kernel32.GetLastError
GetLastError.argtypes = []
GetLastError.restype = DWORD

advapi32 = WinDLL("advapi32", use_last_error=True)
OpenProcessToken = advapi32.OpenProcessToken
OpenProcessToken.argtypes = [HANDLE, DWORD, PHANDLE]
OpenProcessToken.restype = BOOL
GetTokenInformation = advapi32.GetTokenInformation
# GetTokenInformation.argtypes = [HANDLE, TOKEN_INFORMATION_CLASS, LPVOID, DWORD, PDWORD]
GetTokenInformation.restype = BOOL


class SID_IDENTIFIER_AUTHORITY(Structure):
    # noinspection PyTypeChecker
    _fields_ = [
        ('Value', BYTE * 6),
    ]

    def __repr__(self):
        return "<SID_IDENTIFIER_AUTHORITY: %s" % (":".join(str(v) for v in self.Value))


PSID_IDENTIFIER_AUTHORITY = POINTER(SID_IDENTIFIER_AUTHORITY)


class SID(Structure):
    # noinspection PyTypeChecker
    _fields_ = [
        ('Revision', BYTE),
        ('SubAuthorityCount', BYTE),
        ('IdentifierAuthority', SID_IDENTIFIER_AUTHORITY),
        ('SubAuthority', DWORD * 16),
    ]

    def __repr__(self):
        subs = []
        for i in range(self.SubAuthorityCount):
            subs.append(self.SubAuthority[i])
        return "<SID: Revision:%i, SubAuthorityCount:%i, IdentifierAuthority:%s, SubAuthority:%s>" % (
            self.Revision, self.SubAuthorityCount, self.IdentifierAuthority, subs,
        )


PSID = POINTER(SID)


class ACL(Structure):
    _fields_ = [
        ('AclRevision', BYTE),
        ('Sbz1', BYTE),
        ('AclSize', WORD),
        ('AceCount', WORD),
        ('Sbz2', WORD),
    ]


PACL = POINTER(ACL)


class SECURITY_DESCRIPTOR(Structure):
    SECURITY_DESCRIPTOR_CONTROL = WORD
    _fields_ = [
        ('Revision', c_ubyte),
        ('Sbz1', c_ubyte),
        ('Control', SECURITY_DESCRIPTOR_CONTROL),
        ('Owner', PSID),
        ('Group', PSID),
        ('Sacl', PACL),
        ('Dacl', PACL),
    ]

    def __repr__(self):
        def c(v):
            return v.contents if v else 0

        return "<SECURITY_DESCRIPTOR at %#x: Revision:%i, Sbz1:%i, Control:%s, Owner:%s, Group:%s, Sacl=%s, Dacl=%s>" % (
            addressof(self), self.Revision, self.Sbz1, self.Control, c(self.Owner), c(self.Group), c(self.Sacl),
            c(self.Dacl))


PSECURITY_DESCRIPTOR = POINTER(SECURITY_DESCRIPTOR)

InitializeSecurityDescriptor = advapi32.InitializeSecurityDescriptor
SetSecurityDescriptorOwner = advapi32.SetSecurityDescriptorOwner
# don't set this argtypes, or you will get mysterious segfaults:
# SetSecurityDescriptorOwner.argtypes = [SECURITY_DESCRIPTOR, PSID, BOOL]
SetSecurityDescriptorOwner.restype = BOOL
SetSecurityDescriptorGroup = advapi32.SetSecurityDescriptorGroup
# don't set this argtypes, or you will get mysterious segfaults:
# SetSecurityDescriptorGroup.argtypes = [SECURITY_DESCRIPTOR, PSID, BOOL]
SetSecurityDescriptorGroup.restype = BOOL
SetSecurityDescriptorDacl = advapi32.SetSecurityDescriptorDacl
SetSecurityDescriptorDacl.argtypes = [PSECURITY_DESCRIPTOR, BOOL, PACL, BOOL]
SetSecurityDescriptorDacl.restype = BOOL
SetSecurityDescriptorSacl = advapi32.SetSecurityDescriptorSacl
SetSecurityDescriptorSacl.argtypes = [PSECURITY_DESCRIPTOR, BOOL, PACL, BOOL]
SetSecurityDescriptorSacl.restype = BOOL

AllocateAndInitializeSid = advapi32.AllocateAndInitializeSid
AllocateAndInitializeSid.argtypes = [PSID_IDENTIFIER_AUTHORITY, BYTE, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD,
                                     DWORD, PSID]
AllocateAndInitializeSid.restype = BOOL
FreeSid = advapi32.FreeSid
FreeSid.argtypes = [PSID]
FreeSid.restype = PVOID
InitializeAcl = advapi32.InitializeAcl
InitializeAcl.argtypes = [PACL, DWORD, DWORD]
InitializeAcl.restype = BOOL
GetLengthSid = advapi32.GetLengthSid
GetLengthSid.argtypes = [PSID]
GetLengthSid.restype = DWORD
CreateWellKnownSid = advapi32.CreateWellKnownSid
CreateWellKnownSid.argtypes = [WORD, PSID, PSID, PDWORD]
CreateWellKnownSid.restype = BOOL


class ACE_HEADER(Structure):
    _fields_ = [
        ('AceType', BYTE),
        ('AceFlags', BYTE),
        ('SidStart', WORD),
    ]


class ACCESS_ALLOWED_ACE(Structure):
    _fields_ = [
        ('Header', ACE_HEADER),
        ('Mask', DWORD),
        ('SidStart', DWORD),
    ]


AddAccessAllowedAce = advapi32.AddAccessAllowedAce
# AddAccessAllowedAce.argtypes = [PACL, DWORD, DWORD, PSID]
AddAccessAllowedAce.restype = BOOL


class TOKEN_USER(Structure):
    _fields_ = [
        ('SID', PSID),
        ('ATTRIBUTES', DWORD),
    ]


class TOKEN_PRIMARY_GROUP(Structure):
    _fields_ = [
        ('PrimaryGroup', PSID),
    ]



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.server.keyboard_config_base import KeyboardConfigBase
from xpra.log import Logger

log = Logger("keyboard", "osx")


class KeyboardConfig(KeyboardConfigBase):

    def __repr__(self):
        return "osx.KeyboardConfig"

    def get_current_mask(self):
        return []

    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool,
                       modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
        global KEYCODES
        keycode = KEYCODES.get(keyname, -1)
        if keycode == -1:
            keycode = KEYCODES.get(keyname.upper(), -1)
        log("get_keycode%s=%s, %s",
            (client_keycode, keyname, pressed, modifiers, keyval, keystr, group), keycode, group)
        return keycode, group


# we currently assume that all key events are sent using X11 names,
# so we need to translate them to osx keys
# http://x86osx.com/bbs/c_data/pds_comment/MacintoshToolboxEssentials.pdf
KEYCODES: dict[str, int] = {
    # Standardkeys
    "A": 0,
    "B": 11,
    "C": 8,
    "D": 2,
    "E": 14,
    "F": 3,
    "G": 5,
    "H": 4,
    "I": 34,
    "J": 38,
    "K": 40,
    "L": 37,
    "M": 46,
    "N": 45,
    "O": 31,
    "P": 35,
    "Q": 12,
    "R": 15,
    "S": 1,
    "T": 17,
    "U": 32,
    "V": 9,
    "W": 13,
    "X": 7,
    "Y": 16,
    "Z": 6,
    "1": 18,
    "2": 19,
    "3": 20,
    "4": 21,
    "5": 23,
    "6": 22,
    "7": 26,
    "8": 28,
    "9": 25,
    "0": 29,
    "minus": 27,
    "plus": 69,  # KeypadPlus
    "bracketleft": 33,
    "bracketright": 30,
    "semicolon": 41,
    "apostrophe": 39,
    "comma": 43,
    "period": 47,
    "slash": 44,
    "grave": 50,
    "backslash": 42,
    "BackSpace": 51,
    "Escape": 53,
    "Return": 36,
    "Tab": 48,
    "Caps_Lock": 57,
    "space": 49,
    # Numeric pad keys:
    "KP_1": 83,
    "KP_2": 84,
    "KP_3": 85,
    "KP_4": 86,
    "KP_5": 87,
    "KP_6": 88,
    "KP_7": 89,
    "KP_8": 91,
    "KP_9": 92,
    "KP_0": 82,
    "KP_Decimal": 65,
    # "equal" : 24,
    "equal": 81,
    "KP_Divide": 75,
    "KP_Multiply": 67,
    "KP_Subtract": 78,
    "KP_Add": 69,
    "clear": 71,
    "KP_Enter": 76,
    # Navigation keys:
    "Down": 125,
    "Up": 126,
    "Right": 124,
    "Left": 123,
    "Prior": 116,  # Page Up
    "Next": 121,  # Page Down
    "Home": 115,
    "End": 119,
    "Insert": 114,  # help
    "Delete": 117,
    # Function keys:
    "F1": 122,
    "F2": 120,
    "F3": 99,
    "F4": 119,
    "F5": 96,
    "F6": 57,
    "F7": 98,
    "F8": 100,
    "F9": 101,
    "F10": 109,
    "F11": 103,
    "F12": 111,
    "F13": 105,
    "F14": 107,
    "F15": 113,
    "F16": 106,
    "F17": 64,
    "F18": 79,
    "F19": 80,
    "F20": 90,
    # Modifier keys:
    "Shift_L": 56,
    "Shift_R": 60,
    "Control_L": 59,
    "Control_R": 62,
    "Alt_L": 58,  # option
    "Alt_R": 61,
    "Super_L": 55,  # apple
}

log("KEYCODES: %s", KEYCODES)
